
__version__ = "0.3.3"

from ._widgets import AnnotatorWidget, InterpolationWidget, MinimalSurfaceWidget
from packaging import version
from ._napari_version import NAPARI_VERSION

if NAPARI_VERSION >= version.parse("0.4.15"):
    from ._widgets import ListWidgetBB
    __all__ = ["AnnotatorWidget", "InterpolationWidget", "MinimalSurfaceWidget", "ListWidgetBB", "NAPARI_VERSION"]
else:
    __all__ = ["AnnotatorWidget", "InterpolationWidget", "MinimalSurfaceWidget", "NAPARI_VERSION"]




import warnings
import os
import yaml
from qtpy.QtWidgets import QWidget
from qtpy.QtCore import QStandardPaths
from typing import Union
from traceback import print_exc

GETTER_FUN_NAMES = ["isChecked", "value", "text", "currentText"]
SETTER_FUN_NAMES = ["setChecked", "setValue", "setText", "setCurrentText"]

__location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__)))
default_settings_path = os.path.join(__location__, "default_widget_values.yaml")


class UniqueDict(dict):
    def __setitem__(self, key, value):
        if key in self:
            raise ValueError("key %s already taken (=%s)" % (key, self[key]))
        super().__setitem__(key, value)


class PersistentWidgetState:
    def __init__(self):
        config_folder = QStandardPaths.writableLocation(QStandardPaths.ConfigLocation)
        self._config_path = os.path.join(config_folder, "nd_annotator_config.yaml")
        self._state: UniqueDict = UniqueDict()
        if os.path.exists(self._config_path):
            try:
                with open(self._config_path, "r") as f:
                    self._state = yaml.safe_load(f)
            except Exception:
                print_exc()
        else:
            try:
                with open(self._config_path, "w") as new_file, open(default_settings_path, "r") as def_file:
                    new_file.write(def_file.read())
                    def_file.seek(0)
                    self._state = yaml.safe_load(def_file)
            except Exception:
                print_exc()

    def store_multiple_state(self, parent_name: str, widget_id_map: dict):
        widget_state = self[parent_name]
        for id_, widget in widget_id_map.items():
            self.store_state(widget_state, id_, widget)

    def store_state(self, parent: Union[str, dict], widget_id: str, widget: QWidget):
        if type(parent) is str:
            widget_state = self[parent]
        else:
            widget_state = parent
        getter_fun = None
        for fun_name in GETTER_FUN_NAMES:
            if hasattr(widget, fun_name):
                getter_fun = getattr(widget, fun_name)
                break
        if getter_fun is None:
            warnings.warn("Cannot get current value of %s (id: %s):"
                          " object type should define one of (%s)"
                          % (widget, widget_id, ", ".join(GETTER_FUN_NAMES)))
            return
        widget_state[widget_id] = getter_fun()

    def load_multiple_state(self, parent_name: str, widget_id_map: dict):
        widget_state = self._state[parent_name]
        for id_, widget in widget_id_map.items():
            self.load_state(widget_state, id_, widget)

    def load_state(self, parent: Union[str, dict], widget_id: str, widget: QWidget):
        if type(parent) is str:
            widget_state = self[parent]
        else:
            widget_state = parent
        if widget_id not in widget_state:
            warnings.warn("id '%s' not found in config file" % widget_id)
            return
        setter_fun = None
        for fun_name in SETTER_FUN_NAMES:
            if hasattr(widget, fun_name):
                setter_fun = getattr(widget, fun_name)
                break
        if setter_fun is None:
            warnings.warn("Couldn't load value for widget %s (id: %s)"
                          " object type should define one of (%s)"
                          % (widget, widget_id, ", ".join(SETTER_FUN_NAMES)))
            return
        try:
            setter_fun(widget_state[widget_id])
        except TypeError:
            print_exc()

    def __getitem__(self, item):
        if item not in self._state:
            self._state[item] = UniqueDict()
        return self._state[item]

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(PersistentWidgetState, cls).__new__(cls)
        return cls.instance

    def save_state(self):
        state = dict()
        for k, v in self._state.items():
            state[k] = dict(v)
        try:
            with open(self._config_path, "w") as f:
                yaml.dump(state, f)
        except Exception as e:
            print("Couldn't save plugin state due to the following error:")
            print_exc()

    def __del__(self):
        self.save_state()


class PersistentWidget(QWidget):
    _widget_count = 0

    def __init__(self, id_: str = None, **kwargs):
        super().__init__(**kwargs)
        self._annotator_state = PersistentWidgetState()
        self._stored_widgets = dict()
        self._id = id_
        self.destroyed.connect(lambda o: self.on_destroy(self._annotator_state, self._id, self._stored_widgets))
        PersistentWidget._widget_count += 1

    def set_stored_widgets(self, widget_ids: Union[list, dict]):
        self._stored_widgets.clear()
        if type(widget_ids) is list:
            for id_ in widget_ids:
                self._stored_widgets[id_] = getattr(self, id_)
        else:
            for id_, widget in widget_ids.items():
                self._stored_widgets[id_] = widget
        self._annotator_state.load_multiple_state(self._id, self._stored_widgets)

    def add_stored_widget(self, id_, widget=None):
        if widget is None:
            widget = getattr(self, id_)
        self._stored_widgets[id_] = widget
        self._annotator_state.load_state(self._id, id_, widget)

    @staticmethod
    def on_destroy(state, id_, widgets):
        PersistentWidget._widget_count -= 1
        if len(widgets) > 0 and id_ is not None:
            state.store_multiple_state(id_, widgets)
        if PersistentWidget._widget_count == 0:
            state.save_state()
        elif PersistentWidget._widget_count < 0:
            warnings.warn("negative 'PersistentWidget._widget_count': %d" % PersistentWidget._widget_count)



import json
from .util import getReconMethod

class Settings:

    def __init__(self, filename=None, max_iterations=20, debug_mode=False, n_points=1000, resolution_multiplier=1000,
                 iteration_multiplier=10, smooth_reparametrization=False, reconstruction_method="Skip reconstruction",
                 gradient_iterations=100, alpha=0.01, lambd=1.):
        if filename is not None:
            with open(filename) as json_file:
                data = json.load(json_file)
            self.exportDict = data

            # general settings
            self.maxIter = (data['general'])['maxIter']
            self.debug = (data['general'])['debugMode']

            # contour settings
            self.nPoi = (data['contours'])['nPoi']
            self.resMultiplier = (data['contours'])['resMultiplier']

            # reparameterization settings
            self.iterations = self.nPoi*(data['reparameterization'])['iterationMultiplier']
            self.smoothParam = (data['reparameterization'])['smoothReparam']

            # reconstrutcion settings
            self.reconText = (data['reconstruction'])['reconMethod']
            self.reconMethod = getReconMethod((data['reconstruction'])['reconMethod'])
            self.gradientIterations = (data['reconstruction'])['iterations']
            self.alpha = (data['reconstruction'])['alpha']
            self.lambd = (data['reconstruction'])['lambda']

            # export settings
            self.exportCsv = (data['export'])['exportCsv']
            self.exportName = (data['export'])['exportName']
        else:
            # general settings
            self.maxIter = max_iterations
            self.debug = debug_mode

            # contour settings
            self.nPoi = n_points
            self.resMultiplier = resolution_multiplier

            # reparameterization settings
            self.iterations = self.nPoi * iteration_multiplier
            self.smoothParam = smooth_reparametrization

            # reconstrutcion settings
            self.reconText = reconstruction_method
            self.reconMethod = getReconMethod(reconstruction_method)
            self.gradientIterations = gradient_iterations
            self.alpha = alpha
            self.lambd = lambd

    def update(self, category, name, value):
        print("Updating initial entry in category "+category+", name "+name+": "+str((self.exportDict[category])[name])+"...")
        (self.exportDict[category])[name] = value
        print("Updated entry in category "+category+", name "+name+": "+str((self.exportDict[category])[name])+"!")
    
    # update the settings variables according to the exportDict
    def updateVariables(self):
        data = self.exportDict
        
        # general settings
        self.maxIter = (data['general'])['maxIter']
        self.debug = (data['general'])['debugMode']

        # contour settings
        self.nPoi = (data['contours'])['nPoi']
        self.resMultiplier = (data['contours'])['resMultiplier']

        # reparameterization settings
        self.iterations = self.nPoi*(data['reparameterization'])['iterationMultiplier']
        self.smoothParam = (data['reparameterization'])['smoothReparam']

        # reconstrutcion settings
        self.reconText = (data['reconstruction'])['reconMethod']
        self.reconMethod = getReconMethod((data['reconstruction'])['reconMethod'])
        self.gradientIterations = (data['reconstruction'])['iterations']
        self.alpha = (data['reconstruction'])['alpha']
        self.lambd = (data['reconstruction'])['lambda']

        # export settings
        self.exportCsv = (data['export'])['exportCsv']
        self.exportName = (data['export'])['exportName']



import warnings

from magicgui.widgets import PushButton
from time import sleep
from copy import deepcopy
from collections.abc import Iterable
from packaging import version
from threading import Thread
import numpy as np
import itertools
import colorsys
from random import shuffle, seed

#image processing modules
import cv2
import skimage.measure, skimage.morphology
from scipy.ndimage import gaussian_filter, zoom

#UI
from magicclass import magicclass, MagicTemplate, field, vfield, set_design
from magicclass.serialize import serialize
from magicclass.widgets import FreeWidget

from qtpy.QtCore import Signal, Qt, QObject, QThread
from qtpy.QtWidgets import (
    QVBoxLayout,
    QLabel,
    QCheckBox,
    QPushButton,
    QFrame,
    QWidget,
    QApplication,
    QMessageBox
)

import napari
from napari.layers import Image, Labels, Points, Shapes, Layer
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.notifications import show_info
from napari.layers._source import layer_source


from ._utils import WidgetWithLayerList, ProgressWidget, QDoubleSlider, QSymmetricDoubleRangeSlider, ScriptExecuteWidget
from ._utils.callbacks import keep_layer_on_top
from .._helper_functions import layer_dims_displayed
from ._utils.collapsible_widget import CollapsibleContainerGroup, correct_container_size
from .minimal_contour_widget import MinimalContourWidget, delay_function
from .._napari_version import NAPARI_VERSION


try:
    from napari_bbox import BoundingBoxLayer
except ImportError:
    BoundingBoxLayer = None

try:
    import minimal_surface
except ImportError:
    minimal_surface = None


GRADIENT = "Gradient"
CUSTOM = "Custom"
FEATURE_KEYS = [GRADIENT, CUSTOM]
# FEATURES = {
#     ASSYM_GRADIENT: GRADIENT_BASED,
#     SYM_GRADIENT: INTENSITY_BASED,
#     HIGH_INTENSITY: INTENSITY_BASED,
#     LOW_INTENSITY: INTENSITY_BASED,
#     CUSTOM: INTENSITY_BASED
# }

CAD_BLURRING = "Edge preserving" #"Curvature Anisotropic Diffusion"
GAUSSIAN_BLURRING = "Gaussian"

def pts_2_bb(p1, p2, image_size, scale=1.):
    center = (p1 + p2) / 2
    size = np.divide(np.sqrt((np.multiply(p1 - p2, scale) ** 2).sum()), scale) + 10
    bb = np.asarray(
        np.where(list(itertools.product((False, True), repeat=3)), center + size / 2, center - size / 2))
    bb = np.clip(bb, 0, np.asarray(image_size) - 1)
    return bb


def get_bb_corners(bb):
    return np.concatenate([bb.min(0, keepdims=True), bb.max(0, keepdims=True)])


def bb_2_slice(bb):
    bb_corners = get_bb_corners(bb).round().astype(int)
    return tuple(slice(bb_corners[0, i], bb_corners[1, i]) for i in range(3))


def color_to_hex_string(rgb_tuple):
    return '#%.2X%.2X%.2X' % rgb_tuple


def generate_label_colors(n):
    rgb_tuples = [colorsys.hsv_to_rgb(x * 1.0 / n, 1., 1.) for x in range(n)]
    rgb_tuples = [(int(rgb_tuples[i][0] * 255), int(rgb_tuples[i][1] * 255), int(rgb_tuples[i][2] * 255)) for i in
                  range(n)]
    seed(0)
    shuffle(rgb_tuples)
    return rgb_tuples


def generate_sphere_mask(shape):
    ndim = len(shape)
    r = max(shape) // 2
    mask = skimage.morphology.ball(r)
    shape_diffs = tuple(shape[i] - mask.shape[i] for i in range(ndim))
    crop_idx = tuple(
        slice(max(0, round(-shape_diffs[i] / 2 + 0.1)), mask.shape[i] + min(0, round(shape_diffs[i] / 2 + 0.1))) for
        i in range(ndim))
    mask = mask[crop_idx]
    mask = np.pad(mask, [(max(0, round(shape_diffs[i] / 2)), -min(0, round(-shape_diffs[i] / 2 + 0.1))) for i in
                         range(ndim)])
    assert mask.shape == shape
    return mask


class ColorPairsCallback:
    def __init__(self, n_colors=50):
        self.prev_selection = set()
        self.prev_len = 0
        self.prev_data = None
        color_cycle = list(
            map(lambda color_tuple: color_to_hex_string(color_tuple), generate_label_colors(n_colors)))
        self.color_cycle = itertools.cycle(color_cycle)
        self._last_color = None

    def __call__(self, event):
        if event.type != "data" or getattr(event, "action", "add") not in ["add", "added"]:
            return
        points_layer = event.source
        if len(points_layer.selected_data) > 1:
            points_layer.selected_data = self.prev_selection
            points_layer.refresh()
            return
        self.prev_selection = points_layer.selected_data.copy()
        if len(points_layer.data) > self.prev_len:
            if len(points_layer.data) % 2 == 1:
                self._last_color = next(self.color_cycle)
                # points_layer.current_face_color = new_color
            points_layer.face_color[-1] = transform_color(self._last_color)
            self.prev_data = points_layer.data.copy()
        elif len(points_layer.data):
            removed_idx = np.squeeze(np.argwhere(
                np.all(np.any(~np.equal(self.prev_data[np.newaxis], points_layer.data[:, np.newaxis]), -1),
                       axis=0)))
            if removed_idx.ndim == 0:
                removed_idx = int(removed_idx)
                points_layer.selected_data.clear()
                if removed_idx % 2 == 0 and removed_idx < len(points_layer.data):
                    points_layer.selected_data.add(removed_idx)
                elif removed_idx % 2 == 1:
                    points_layer.selected_data.add(removed_idx - 1)
                points_layer.remove_selected()
        points_layer.refresh()
        self.prev_data = points_layer.data
        self.prev_len = len(points_layer.data)


class EstimationWorker(QObject):
    image_data_received = Signal(str, "PyQt_PyObject", "PyQt_PyObject")
    mask_data_received = Signal("PyQt_PyObject", "PyQt_PyObject")
    remove_layer = Signal(str)
    layer_invalidated = Signal(str)
    all_done = Signal()
    annotation_needed_signal = Signal("PyQt_PyObject", "PyQt_PyObject")
    slice_annotations_done = Signal(int)
    object_annotated = Signal(int)
    MANUAL_ANNOTATION = "Manual"
    THRESHOLD_ANNOTATION = "Threshold"
    MINIMAL_CONTOUR_ANNOTATION = "Minimal contour"
    ANNOTATION_METHODS = [MANUAL_ANNOTATION, THRESHOLD_ANNOTATION, MINIMAL_CONTOUR_ANNOTATION]

    def __init__(self, viewer, minimal_contour_widget: MinimalContourWidget, minimal_surface_widget, parent=None):
        super().__init__(parent)
        self.viewer = viewer
        self.minimal_contour_widget = minimal_contour_widget
        self.minimal_surface_widget = minimal_surface_widget
        self.estimators = None
        self.bboxes = None
        self.imgs = None
        self.phis = None
        self.pt_pairs = None
        self.stop_requested = False
        self.center_mat = None
        self.rotation_matrix = None
        self.translation_matrix = None
        self.phi = None
        self.image = None
        self.points = None
        self.use_correction = None
        self.beta = None
        self.alpha = None
        self.n_iter = None
        self.use_meeting_plane_points = None
        self.done_pressed = False
        self.blur_func = None
        self.init_slice_annotation_fun = None
        self.finish_slice_annotation_fun = None
        self.bb_layer = None
        self.slice_image_layer = None
        self.slice_labels_layer = None
        self.slice_annotation = None
        self.manual_annotation_done = False
        self.z_scale = None
        self.use_gradient = None

        self.prev_ndisplay = None
        self.prev_camera_center = None
        self.prev_camera_zoom = None
        self.prev_camera_angles = None

        #create annotation dialogs
        self.manual_annotation_dialog = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(QLabel("Annotate image, then click 'Done'"))
        annotation_done_button = QPushButton("Done")
        layout.addWidget(annotation_done_button)
        def on_done_pressed():
            self.done_pressed = True
        annotation_done_button.clicked.connect(self.finish_slice_annotation)
        annotation_done_button.clicked.connect(on_done_pressed)
        self.manual_annotation_dialog.setLayout(layout)
        self.manual_annotation_dialog.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.Dialog)
        self.manual_annotation_dialog.setAttribute(Qt.WA_ShowWithoutActivating)

        self.threshold_annotation_dialog = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(QLabel("Threshold:"))
        threshold_slider = QDoubleSlider(Qt.Horizontal)
        threshold_slider.valueChanged.connect(self.threshold_value_changed)
        layout.addWidget(threshold_slider)
        annotation_done_button = QPushButton("Done")
        layout.addWidget(annotation_done_button)
        annotation_done_button.clicked.connect(self.finish_slice_annotation)
        annotation_done_button.clicked.connect(on_done_pressed)
        self.threshold_annotation_dialog.setLayout(layout)
        self.annotation_needed_signal.connect(self.begin_slice_annotation)
        self.slice_annotation_methods = {EstimationWorker.MANUAL_ANNOTATION: self.manual_annotation, EstimationWorker.THRESHOLD_ANNOTATION: self.threshold_annotation, EstimationWorker.MINIMAL_CONTOUR_ANNOTATION: self.minimal_contour_annotation}
        self.finish_annotation_methods = {EstimationWorker.MANUAL_ANNOTATION: self.finish_manual_annotation, EstimationWorker.THRESHOLD_ANNOTATION: self.finish_threshold_annotation, EstimationWorker.MINIMAL_CONTOUR_ANNOTATION: self.finish_minimal_contour_annotation}

    def begin_slice_annotation(self, image_slice, distance_map_slice):
        self.slice_image_layer = self.viewer.add_image(image_slice)
        self.slice_labels_layer = self.viewer.add_labels(np.zeros_like(image_slice, dtype=int))
        self.prev_ndisplay = self.viewer.dims.ndisplay
        self.prev_camera_center = self.viewer.camera.center
        self.prev_camera_zoom = self.viewer.camera.zoom
        self.prev_camera_angles = self.viewer.camera.angles
        self.viewer.dims.ndisplay = 2
        visible_extent = self.slice_labels_layer.extent.world[:, layer_dims_displayed(self.slice_labels_layer)]
        self.viewer.camera.center = visible_extent.mean(0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.camera.zoom = np.min(np.divide(self.viewer._canvas_size, visible_extent.max(0)-visible_extent.min(0)))*0.95
        self.slice_labels_layer.brush_size = 1
        self.slice_labels_layer.mode = "paint"
        self.minimal_surface_widget.call_button.enabled = True
        self.init_slice_annotation_fun()

    def finish_slice_annotation(self):
        self.finish_slice_annotation_fun()
        self.minimal_surface_widget.call_button.enabled = False
        if self.slice_labels_layer is not None:
            self.slice_annotation = self.slice_labels_layer.data > 0
        if self.slice_image_layer in self.viewer.layers:
            self.viewer.layers.remove(self.slice_image_layer)
        if self.slice_labels_layer in self.viewer.layers:
            self.viewer.layers.remove(self.slice_labels_layer)
        if self.prev_ndisplay is not None:
            self.viewer.dims.ndisplay = self.prev_ndisplay
        if self.prev_ndisplay is not None:
            self.viewer.camera.center = self.prev_camera_center
        if self.prev_camera_zoom is not None:
            self.viewer.camera.zoom = self.prev_camera_zoom
        if self.prev_camera_angles is not None:
            self.viewer.camera.angles = self.prev_camera_angles
        self.slice_labels_layer = None
        self.slice_image_layer = None

    def manual_annotation(self):
        self.manual_annotation_dialog.show()

    def finish_manual_annotation(self):
        self.manual_annotation_dialog.hide()

    def threshold_annotation(self):
        threshold_slider = self.threshold_annotation_dialog.children()[2]
        threshold_slider.setMinimum(self.slice_image_layer.data.min())
        threshold_slider.setMaximum(self.slice_image_layer.data.max())
        threshold_slider.setValue((threshold_slider.minimum()+threshold_slider.maximum())/2)
        self.threshold_annotation_dialog.show()

    def threshold_value_changed(self, val):
        self.slice_labels_layer.data = self.slice_image_layer.data >= val

    def finish_threshold_annotation(self):
        self.threshold_annotation_dialog.hide()

    def minimal_contour_annotation(self):
        self.viewer.layers.selection.select_only(self.minimal_contour_widget.anchor_points)
        self.minimal_contour_widget.image_layer = self.slice_image_layer
        self.minimal_contour_widget.labels_layer = self.slice_labels_layer
        self.minimal_contour_widget.anchor_points.mode = "add"
        self.manual_annotation_dialog.show()

    def finish_minimal_contour_annotation(self):
        self.manual_annotation_dialog.hide()

    def initialize(self, image, phi_func, points, use_correction, beta, alpha, n_iter, use_meeting_plane_points, slice_annotation_method="Manual", blur_func=None, z_scale=1., use_gradient=True):
        self.image = image
        self.phi_func = phi_func
        self.points = points
        self.use_correction = use_correction
        self.beta = beta
        self.alpha = alpha
        self.n_iter = n_iter
        self.use_meeting_plane_points = use_meeting_plane_points
        self.blur_func = blur_func
        self.init_slice_annotation_fun = self.slice_annotation_methods[slice_annotation_method]
        self.finish_slice_annotation_fun = self.finish_annotation_methods[slice_annotation_method]
        self.z_scale = z_scale
        self.use_gradient = use_gradient

    def hook_callbacks(self, idx):
        estimator = self.estimators[idx]
        # estimator.hook_stage_data_init_event(
        #     minimal_surface.AREA_EIKONAL_STAGE,
        #     lambda arr, idx: self.data_initializer(
        #         "Area Eikonal %d%s" % (0, postscript),
        #         0,
        #         {
        #             'colormap': "plasma",
        #             # 'translate': offset,
        #             "visible": False
        #         }
        #     )(arr, idx)
        # )
        # estimator.hook_stage_iteration_event(minimal_surface.AREA_EIKONAL_STAGE,
        #                                      lambda idx: self.data_updater("Area Eikonal %d%s" % (0, postscript))(
        #                                          idx))
        # estimator.hook_stage_finished_event(minimal_surface.AREA_EIKONAL_STAGE,
        #                                     lambda: self.data_finalizer("Area Eikonal %d%s" % (0, postscript))())
        # estimator.hook_stage_data_init_event(
        #     minimal_surface.AREA_EIKONAL_STAGE,
        #     lambda arr, idx: self.data_initializer(
        #         "Area Eikonal %d%s" % (1, postscript),
        #         1,
        #         {
        #             'colormap': "plasma",
        #             # 'translate': offset,
        #             "visible": False
        #         }
        #     )(arr, idx)
        # )
        # estimator.hook_stage_iteration_event(minimal_surface.AREA_EIKONAL_STAGE,
        #                                      lambda idx: self.data_updater("Area Eikonal %d%s" % (1, postscript))(
        #                                          idx))
        # estimator.hook_stage_finished_event(minimal_surface.AREA_EIKONAL_STAGE,
        #                                     lambda: self.data_finalizer("Area Eikonal %d%s" % (1, postscript))())
        # self.hook_callbacks(estimator, minimal_surface.AREA_EIKONAL_STAGE, "Area Eikonal 0", {'colormap': "plasma", 'translate': offset, "visible": False}, 0)
        # self.hook_callbacks(estimator, minimal_surface.AREA_EIKONAL_STAGE, "Area Eikonal 1", {'colormap': "plasma", 'translate': offset, "visible": False}, 1)
        # self.hook_callbacks(estimator, minimal_surface.ROTATED_AREA_EIKONAL_STAGE, "Rotated Area Eikonal", {'colormap': "plasma"}, 0)
        # estimator.hook_stage_data_init_event(minimal_surface.PLANE_PHASEFIELD_STAGE,
        #                                      lambda arr, idx: print("plane phasefield eikonal init", arr.shape, idx))
        # estimator.hook_stage_data_init_event(
        #     minimal_surface.PLANE_PHASEFIELD_STAGE,
        #     lambda arr, idx: self.data_initializer("Plane PhaseField%s" % postscript,
        #                                            layer_args={
        #                                                'colormap': "plasma",
        #                                                'translate': offset + self.translation_matrix +
        #                                                             np.asarray([0, 0, self.center_mat[
        #                                                                 -1]]) @ self.rotation_matrix.T,
        #                                                'rotate': self.rotation_matrix,
        #                                                'opacity': 0.6,
        #                                                "visible": False
        #                                            }
        #                                            )(arr, idx)
        # )
        # estimator.hook_stage_iteration_event(minimal_surface.PLANE_PHASEFIELD_STAGE,
        #                                      lambda idx: self.data_updater("Plane PhaseField%s" % postscript)(idx))
        # estimator.hook_stage_finished_event(minimal_surface.PLANE_PHASEFIELD_STAGE,
        #                                     lambda: self.data_finalizer("Plane PhaseField%s" % postscript)())

        # estimator.hook_stage_data_init_event(minimal_surface.TRANSPORT_FUNCTION_STAGE,
        #                                      lambda arr, idx: print("plane phasefield eikonal init", arr.shape, idx))

        # estimator.hook_stage_data_init_event(
        #     minimal_surface.TRANSPORT_FUNCTION_STAGE,
        #     lambda arr, idx: self.data_initializer("Transport Function%s" % postscript,
        #                                            0,
        #                                            {
        #                                                'colormap': "turbo",
        #                                                'translate': offset + self.translation_matrix,
        #                                                'rotate': self.rotation_matrix,
        #                                                'opacity': 0.5,
        #                                                'rendering': "iso",
        #                                                "iso_threshold": 0.
        #                                            })(arr, idx)
        # )
        # # estimator.hook_stage_data_init_event(minimal_surface.TRANSPORT_FUNCTION_STAGE, lambda arr, idx: print("tf init"))
        # estimator.hook_stage_iteration_event(minimal_surface.TRANSPORT_FUNCTION_STAGE,
        #                                      lambda idx: self.data_updater("Transport Function%s" % postscript)(
        #                                          idx))
        # # estimator.hook_stage_iteration_event(minimal_surface.TRANSPORT_FUNCTION_STAGE, lambda idx: print("tf iter %d" % idx))
        # estimator.hook_stage_finished_event(minimal_surface.TRANSPORT_FUNCTION_STAGE,
        #                                     lambda: self.data_finalizer("Transport Function%s" % postscript)())

        def tform_calculated(rotation, translation):
            self.rotation_matrix = np.flip(rotation.copy()).reshape(3, 3)
            self.translation_matrix = np.flip(translation.copy())
        # print("hook_transform_calculated_event")
        # estimator.hook_transform_calculated_event(tform_calculated)
        # print("after hook_transform_calculated_event")

        def center_calculated(center):
            self.center_mat = np.flip(center.copy())
        # print("hook_plane_center_calculated_event")
        # estimator.hook_plane_center_calculated_event(center_calculated)
        # print("after hook_plane_center_calculated_event")

    def run(self):
        import time
        self.minimal_surface_widget.call_button.clicked.disconnect(self.minimal_surface_widget._start_estimation)
        def stop():
            self.stop_requested = True
            self.minimal_surface_widget.call_button.text = "Stopping..."
            self.finish_slice_annotation()
            self.minimal_surface_widget.call_button.enabled = True
        self.minimal_surface_widget.call_button.clicked.connect(stop)
        self.minimal_surface_widget.call_button.text = "Stop"
        self.minimal_surface_widget.call_button.enabled = False

        try:
            start = time.time()

            self.estimators = []
            self.bboxes = []
            self.imgs = []
            self.phis = []
            self.pt_pairs = []
            for i in range(len(self.points) // 2):
                estimator = minimal_surface.MinimalSurfaceCalculator()
                self.estimators.append(estimator)
                estimator.set_initial_plane_calculator(self.segment_initial_slice)
                estimator.set_using_meeting_points(self.use_meeting_plane_points)
                self.hook_callbacks(i)
                bounding_box = pts_2_bb(self.points[2 * i], self.points[2 * i + 1], self.image.shape,
                                        [self.z_scale, 1, 1])
                bounding_box = bounding_box.round().astype(int)
                self.bboxes.append(bounding_box)
                offset = bounding_box.min(0, keepdims=True)
                bb_slice = bb_2_slice(bounding_box)
                point1 = (self.points[2 * i] - offset).reshape(-1)
                point2 = (self.points[2 * i + 1] - offset).reshape(-1)
                point1[2] *= self.z_scale
                point2[2] *= self.z_scale
                self.pt_pairs.append((point1, point2))
                data = self.image[bb_slice]
                data = zoom(data, (self.z_scale, 1, 1))
                data = (data - (min_ := data.min())) / (data.max() - min_)
                if self.blur_func is not None:
                    data = self.blur_func(data)
                if self.use_gradient:
                    phi = sitk.GetArrayFromImage(sitk.GradientMagnitude(sitk.GetImageFromArray(data))).astype(float)
                else:
                    phi = self.phi_func(data)
                alpha = (self.alpha - np.exp(-self.beta))/(1-np.exp(-self.beta))
                phi = alpha + (1-alpha)*np.exp(-self.beta*phi)
                phi = phi/phi.max()
                self.imgs.append(data)
                self.phis.append(phi)
                if self.use_meeting_plane_points:
                    estimator.calc_eikonal_and_transport_init(phi, data, point1, point2, self.use_correction)
                else:
                    estimator.init_transport_slice(phi, point1, point2)
            self.slice_annotations_done.emit(len(self.estimators))

            show_info("Slice annotations finished in %.2f seconds" % (time.time() - start))
            timestamp = time.time()

            for i, (estimator, phi, data, (point1, point2), bounding_box)\
                    in enumerate(zip(self.estimators, self.phis, self.imgs, self.pt_pairs, self.bboxes)):
                if self.stop_requested:
                    show_info("Stopped calculation")
                    break
                if self.bb_layer is not None:
                    self.bb_layer.data = [bounding_box]

                offset = np.clip(bounding_box.min(0), 0, np.asarray(self.image.shape) - 1)
                # start = time.time()
                output = estimator.calculate(phi, data, point1, point2, self.use_correction, self.n_iter)
                segmented = (output >= 0)
                labelled = skimage.measure.label(segmented)
                obj_pixel = np.argwhere(output == output.max())[0]
                obj_label = labelled[tuple(obj_pixel)]
                mask = (labelled == obj_label)
                mask = zoom(mask, (1/self.z_scale, 1, 1), order=0)
                # self.image_data_received.emit("Result", output, {"colormap": "plasma", "translate": offset})
                self.mask_data_received.emit(mask, offset)
                postscript = " - %d" % (i + 1)
                # self.remove_layer.emit("Transport Function")
                self.object_annotated.emit(i)
            if not self.stop_requested:
                show_info("Estimation done in %d" % (time.time()-start))
        finally:
            self.stop_requested = False
            self.all_done.emit()
            self.minimal_surface_widget.call_button.clicked.disconnect(stop)
            self.minimal_surface_widget.call_button.clicked.connect(self.minimal_surface_widget._start_estimation)
            self.minimal_surface_widget.call_button.text = "Run"
            self.minimal_surface_widget.call_button.enabled = True

    def data_initializer(self, name, selected_idx=None, layer_args=None):
        if layer_args is None:
            layer_args = {}

        def initialize(arr, idx):
            try:
                if selected_idx in [None, idx]:
                    self.image_data_received.emit(name, arr, layer_args)
                else:
                    print("%s was not initialized as id (%d) was not %d" % (name, idx, selected_idx))
            except Exception as e:
                ...

        return initialize
        # return lambda arr, idx: print("initializing", name)

    def data_updater(self, name):
        it = [-1]

        def update_viewer(iteration):
            it[0] += 1
            if it[0] % 100 == 0: #TODO undo
                self.layer_invalidated.emit(name)
        return update_viewer

    def data_finalizer(self, name):
        def finalize():
            if name not in self.viewer.layers:
                print("%s not in layer list" % name)
                return
            self.viewer.layers[name].data = np.copy(self.viewer.layers[name].data)

        return finalize

    def segment_initial_slice(self, image_slice, distance_map):
        image_slice = np.squeeze(image_slice)
        distance_map = np.squeeze(distance_map)
        self.annotation_needed_signal.emit(image_slice, distance_map)
        while not self.done_pressed and not self.stop_requested:
            sleep(0.1)
            QApplication.processEvents()
        self.done_pressed = False
        if self.slice_annotation is None:
            raise ValueError("Slice not calculated")
        mask = self.slice_annotation.astype(float) if self.slice_annotation is not None else None
        self.slice_annotation = None
        return mask


class SliderWithCheckbox(FreeWidget):
    def __init__(self, state=False, value=None, min=None, max=None, enabled=True):
        super().__init__("horizontal")
        self.wdt = QWidget()
        layout = QVBoxLayout()
        self._checkbox = QCheckBox()
        layout.addWidget(self._checkbox)
        self._slider = QSymmetricDoubleRangeSlider(Qt.Horizontal)
        if min is not None:
            self.min = min
        if max is not None:
            self.max = max
        self.enabled = enabled
        if value is not None:
            self._slider.setValue(value)
        layout.addWidget(self._slider)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.wdt.setLayout(layout)
        self.set_widget(self.wdt)
        self.checked = state

    @property
    def checked(self):
        return self._checkbox.isChecked()

    @checked.setter
    def checked(self, new_value):
        if new_value == self._checkbox.isChecked():
            return
        self._checkbox.setChecked(new_value)
        self._slider.setEnabled(new_value)
        self.state_changed.emit(self.checked)

    @property
    def state_changed(self):
        return self._checkbox.clicked

    @property
    def value_changed(self):
        return self._slider.valueChanged

    @property
    def value(self):
        return self._slider.value()

    @value.setter
    def value(self, new_value):
        self._slider.setValue(new_value)

    @property
    def max(self):
        return self._slider.maximum()

    @max.setter
    def max(self, new_max):
        self._slider.setMaximum(new_max)

    @property
    def min(self):
        return self._slider.minimum()

    @min.setter
    def min(self, new_min):
        self._slider.setMinimum(new_min)


@magicclass(name="Minimal Surface", widget_type="scrollable", properties={"x_enabled": False})
class _MinimalSurfaceWidget(MagicTemplate):
    image_layer_combobox = field(Image, label="Image")
    labels_layer_combobox = field(Labels, label="Labels")

    @magicclass(widget_type="collapsible", name="Image Features")
    class ImageFeaturesWidget(MagicTemplate):
        ...

    @magicclass(widget_type="collapsible", name="Blurring")
    class BlurringWidget(MagicTemplate):
        def try_blurring(self): ...

    @magicclass(widget_type="collapsible", name="Slice Annotation Method")
    class SliceAnnotationWidget(MagicTemplate):
        ...

    @magicclass
    class ImageSliceWidget(MagicTemplate):
        dim_spinbox = field(int, label="Axis").with_options(max=0)
        position_slider = field(int, label="Position", widget_type="Slider").with_options(max=0)

        def __init__(self):
            super().__init__()
            self._viewer = None
            self._layer = None
            self._slice_layer = None
            self._ctrl_down = False
            self._mouse_drag_callbacks = []
            self._update_slice()
            self._update_widgets_state()
            self._update_max_position()
            self._clipping_widget = None

        @property
        def viewer(self):
            return self._viewer

        @viewer.setter
        def viewer(self, viewer):
            self._viewer = viewer
            viewer.dims.events.ndisplay.connect(self._on_ndisplay_changed)

        @property
        def clipping_widget(self):
            if self._clipping_widget is None:
                raise AttributeError("clipping_widget not initialized")
            return self._clipping_widget

        @clipping_widget.setter
        def clipping_widget(self, clipping_widget):
            if self._clipping_widget is not None:
                raise AttributeError("clipping_widget already initialized")
            self._clipping_widget = clipping_widget

        @property
        def position(self):
            return self.position_slider.value

        @property
        def dimension(self):
            return self.dim_spinbox.value

        @property
        def layer(self):
            return self._layer

        @layer.setter
        def layer(self, new_layer):
            self.clipping_widget.remove_layer(self._layer)
            self._layer = new_layer
            self.clipping_widget.add_layer(self._layer)
            self._update_slice()
            self._update_max_position()
            self._update_max_dim()
            self._update_widgets_state()
            if self._viewer.dims.ndisplay == 3:
                self._create_slice_layer()

        @dim_spinbox.connect
        def _on_dim_changed(self, new_dim):
            self._update_max_position()
            self._update_slice()
            self._update_widgets_state()

        def _update_max_position(self):
            if self.layer is None:
                return
            self.position_slider.max = self.layer.data.shape[self.dimension] - 1

        def _update_max_dim(self):
            if self.layer is None:
                return
            self.dim_spinbox.native.setMaximum(self.layer.data.ndim - 1)

        @position_slider.connect
        def _update_slice(self):
            if self._slice_layer is None:
                return
            dim = self.dimension
            idx = tuple(slice(self.position, self.position + 1) if i == dim else slice(None)
                        for i in range(self.layer.data.ndim))
            self._slice_layer.data = self.layer.data[idx]
            self._slice_layer.translate = [self.position * self.layer.scale[i] if i == dim else 0 for i in
                                           range(self.layer.data.ndim)]

        def _on_mouse_scroll(self, layer, event):
            if "Alt" not in event.modifiers:
                return
            self.position_slider.value = int(self.position - event.delta[0])

        def _update_widgets_state(self):
            is_enabled = self.layer is not None and self._viewer.dims.ndisplay == 3
            self.dim_spinbox.enabled = is_enabled
            self.position_slider.enabled = is_enabled

        def _remove_slice_layer(self):
            if self._slice_layer is not None:
                self._viewer.layers.remove(self._slice_layer)
                self.clipping_widget.remove_layer(self._slice_layer)
                self._slice_layer = None

        def _create_slice_layer(self):
            self._remove_slice_layer()
            if self.layer is None:
                return
            self._slice_layer = self._viewer.add_image(data=np.empty((1, 1, 1)), colormap="red",
                                                       scale=self.layer.scale, blending="additive",
                                                       experimental_clipping_planes=self.layer.experimental_clipping_planes)
            self.clipping_widget.add_layer(self._slice_layer)
            self._slice_layer.mouse_drag_callbacks.extend(self._mouse_drag_callbacks)
            self._slice_layer.mouse_wheel_callbacks.append(self._on_mouse_scroll)
            self._update_slice()
            self._update_widgets_state()
            self._slice_layer.reset_contrast_limits()

        def _on_ndisplay_changed(self):
            if self._viewer.dims.ndisplay == 2:
                self._remove_slice_layer()
            elif self._viewer.dims.ndisplay == 3:
                self._create_slice_layer()
            self._update_widgets_state()

        @set_design(visible=False)
        def add_mouse_drag_callback(self, callback):
            self._mouse_drag_callbacks.append(callback)
            if self._slice_layer is not None:
                self._slice_layer.mouse_drag_callbacks.append(callback)
            return callback

    call_button = field(widget_type=PushButton, name="Run")

    def _start_estimation(self):
        if self.image_layer not in self._viewer.layers:
            warnings.warn("Missing image layer")
            return
        if len(self.points_layer.data) == 0:
            warnings.warn("No points were added")
            return
        if len(self.points_layer.data) == 1:
            warnings.warn("Not enough points")
            return
        if self.labels_layer is None:
            answer = QMessageBox.question(
                self.native,
                "Missing Labels layer",
                "Create Labels layer?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
            )
            if answer == QMessageBox.StandardButton.Yes:
                new_labels = self._viewer.add_labels(
                    np.zeros(self.image_layer.data.shape[:self.image_layer.ndim], dtype=np.uint16))
                self.labels_layer = new_labels
            else:
                return
        elif not np.array_equal(
                self.labels_layer.data.shape[:self.labels_layer.ndim],
                self.image_layer.data.shape[:self.image_layer.ndim]
        ):
            answer = QMessageBox.question(
                self.native,
                "Shape mismatch",
                "Current labels and image layers have different shape."
                " Do you want to create a new labels layer with the appropriate shape?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
            )
            if answer == QMessageBox.StandardButton.Yes:
                new_labels = self._viewer.add_labels(
                    np.zeros(self.image_layer.data.shape[:self.image_layer.ndim], dtype=np.uint16))
                self.labels.layer = new_labels
            else:
                return
        image = self.image_layer
        image_data = image.data
        phi_func = None if (self.used_feature == "Gradient") else self._run_feature_extractor
        points = self.points_layer
        points_data = points.data.copy()
        z_scale = self.z_scale
        alpha = self.alpha
        beta = self.beta
        n_iter = self.iterations
        use_correction = self.use_correction
        use_gradient = self.used_feature == "Gradient"
        use_meeting_plane_points = True  # TODO check self.use_meeting_points_checkbox.isChecked()
        estimation_worker = EstimationWorker(self._viewer, self.minimal_contour_widget, self)
        estimation_worker.initialize(image_data, phi_func, points_data, use_correction, beta, alpha, n_iter, use_meeting_plane_points, self.slice_annotation_method, self._blur_image if self.is_blurring_enabled else None, z_scale, use_gradient)
        estimation_worker.image_data_received.connect(self._add_image)
        estimation_worker.layer_invalidated.connect(self._refresh_layer)
        estimation_worker.mask_data_received.connect(self._mask_received)
        estimation_worker.remove_layer.connect(self._data_remover)
        estimation_worker.slice_annotations_done.connect(self._show_estimation_progress)
        estimation_worker.object_annotated.connect(lambda i: self.estimation_progress_widget.setValue(i+1))
        estimation_thread = QThread(self.native)
        estimation_worker.moveToThread(estimation_thread)
        estimation_worker.all_done.connect(estimation_thread.quit)
        estimation_worker.all_done.connect(lambda: self.estimation_progress_widget.setVisible(False))
        estimation_thread.started.connect(estimation_worker.run)
        estimation_thread.finished.connect(estimation_worker.deleteLater)
        estimation_thread.finished.connect(estimation_thread.deleteLater)
        if BoundingBoxLayer is not None:
            estimation_worker.bb_layer = BoundingBoxLayer(
                [[[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1], [1, 0, 0], [1, 0, 1], [1, 1, 0], [1, 1, 1]]], scale=(z_scale, 1., 1.))
            self._viewer.add_layer(estimation_worker.bb_layer)
            estimation_thread.finished.connect(lambda: self._viewer.layers.remove(estimation_worker.bb_layer))
        estimation_thread.start()
        # viewer.add_image(data)

    @magicclass(widget_type="collapsible")
    class ClippingWidget(MagicTemplate):
        slider_0 = field(SliderWithCheckbox)
        slider_1 = field(SliderWithCheckbox)
        slider_2 = field(SliderWithCheckbox)

        def __init__(self):
            super().__init__()
            self._viewer: napari.Viewer = None
            self._layers = []
            self._ghost_layers = dict()
            self.sliders = []
            self._clipping_planes = []
            self.n_checked = 0

        def __post_init__(self):
            for i, slider in enumerate([self.slider_0, self.slider_1, self.slider_2]):
                slider.name = str(i)
                slider.state_changed.connect(self._on_checked)
                slider.value_changed.connect(self._range_change_handler(i))
                self.sliders.append(slider)
            self._initialize_clipping_planes()

        @property
        def viewer(self):
            return self._viewer

        @viewer.setter
        def viewer(self, viewer):
            self._viewer = viewer
            viewer.dims.events.ndisplay.connect(self._on_ndisplay_changed)

        @property
        def layers(self):
            return list(self._layers)

        @layers.setter
        def layers(self, new_layers):
            new_layers = list(new_layers) if isinstance(new_layers, Iterable) \
                else [new_layers] if new_layers else []
            if self._layers:
                for layer in self.layers:
                    if layer not in new_layers:
                        self.remove_layer(layer)
                    else:
                        new_layers.remove(layer)
            for layer in new_layers:
                self.add_layer(layer)
            self._update_ranges()

        @set_design(visible=False)
        def add_layer(self, layer):
            if layer is None:
                return
            if layer in self.layers:
                return
            self._layers.append(layer)
            self._ghost_layers[layer] = None
            self._update_ghost_layers()
            layer.events.scale.connect(self._update_ranges)
            layer.experimental_clipping_planes = self._clipping_planes
            layer.events.data.connect(self._on_layer_data_change)
            self._update_ranges()

        @set_design(visible=False)
        def remove_layer(self, layer):
            if layer is None:
                return
            if layer not in self.layers:
                return
            self._layers.remove(layer)
            if self._ghost_layers[layer] is not None:
                self.viewer.layers.remove(self._ghost_layers[layer])
            del self._ghost_layers[layer]
            layer.events.scale.disconnect(self._update_ranges)
            layer.events.data.disconnect(self._on_layer_data_change)
            layer.experimental_clipping_planes = []
            self._update_ranges()

        def _on_layer_data_change(self, event):
            layer = event.source
            if layer not in self._ghost_layers:
                print("%s not in ghost layers!" % layer.name if layer else None)
                return
            ghost_layer = self._ghost_layers[event.source]
            if ghost_layer is None:
                return
            ghost_layer.data = event.source.data

        def _range_change_handler(self, idx):
            def on_range_change(value):
                self._clipping_planes[idx * 2]["position"][idx] = value[0]
                self._clipping_planes[idx * 2 + 1]["position"][idx] = value[1]
                self._update_clipping_planes()

            return on_range_change

        def _on_checked(self, value):
            self.n_checked += 1 if value else -1
            self._update_ghost_layers()
            self._update_clipping_planes()

        def _update_ghost_layers(self):
            if self.viewer is None:
                return
            for layer in self._layers:
                ghost = self._ghost_layers[layer]
                if ghost is None and self.n_checked > 0 and self.viewer.dims.ndisplay == 3:
                    data, state, type_str = layer.as_layer_data_tuple()
                    state["name"] = u"\U0001F47B " + layer.name
                    state["opacity"] /= 2.
                    state["experimental_clipping_planes"] = []
                    if NAPARI_VERSION <= version.parse("0.4.16") and "interpolation" in state:
                        interpolation = state["interpolation"]
                        state["interpolation"] = "nearest"
                    with layer_source(parent=layer):
                        new = Layer.create(deepcopy(data), state, type_str)
                    self.viewer.add_layer(new)
                    if NAPARI_VERSION <= version.parse("0.4.16") and "interpolation" in state:
                        new.interpolation = interpolation
                    self.viewer.layers.link_layers([layer, new])
                    self.viewer.layers.unlink_layers([layer, new],
                                                      ["opacity", "experimental_clipping_planes", "blending"])
                    self.viewer.layers.move(self.viewer.layers.index(new), self.viewer.layers.index(layer) + 1)
                    self._ghost_layers[layer] = new
                elif ghost is not None and (self.n_checked == 0 or self.viewer.dims.ndisplay == 2):
                    self.viewer.layers.remove(ghost)
                    self._ghost_layers[layer] = None

        def _remove_ghost_layers(self):
            for layer in self._layers:
                ghost = self._ghost_layers[layer]
                if ghost is not None:
                    self.viewer.layers.remove(ghost)
                    self._ghost_layers[layer] = None

        def _on_ndisplay_changed(self, e):
            if e.value == 2:
                self._remove_ghost_layers()
            else:
                self._update_ghost_layers()

        def _update_ranges(self):
            if not self.layers:
                return
            print("layers:", self.layers)
            nan = np.finfo(float).max
            negnan = np.finfo(float).min

            extent = np.asarray([[nan, nan, nan], [negnan, negnan, negnan]])
            for layer in self.layers:
                layer_extent = layer.extent.world
                if layer.ndim == 2:
                    layer_extent = np.concatenate([np.asarray([[nan], [negnan]]), layer_extent], axis=1)
                extent[0] = np.minimum(extent[0], layer_extent[0])
                extent[1] = np.maximum(extent[1], layer_extent[1])
            for s_idx, slider in enumerate(self.sliders):
                slider.checked = slider.checked or not np.any(np.equal(extent[:, s_idx], np.asarray([nan, negnan])))
                slider.min = extent[0, s_idx]
                slider.max = max(slider.min+1, extent[1, s_idx])
            self._update_clipping_planes()

        def _initialize_clipping_planes(self):
            new_clipping_planes = []
            for s_idx, slider in enumerate(self.sliders):
                start, end = slider.value
                clipping_plane0 = {
                    "position": [start if i == s_idx else 0 for i in range(3)],
                    "normal": [1 if i == s_idx else 0 for i in range(3)]
                }
                new_clipping_planes.append(clipping_plane0)
                clipping_plane1 = {
                    "position": [end if i == s_idx else 0 for i in range(3)],
                    "normal": [-1 if i == s_idx else 0 for i in range(3)]
                }
                new_clipping_planes.append(clipping_plane1)
            self._clipping_planes = new_clipping_planes
            for layer in self.layers:
                layer.experimental_clipping_planes = self._clipping_planes

        def _update_clipping_planes(self):
            new_clipping_planes = []
            for s_idx, slider in enumerate(self.sliders):
                if not slider.checked:
                    continue
                new_clipping_planes.extend(self._clipping_planes[s_idx * 2:s_idx * 2 + 2])
            for layer in self.layers:
                layer.experimental_clipping_planes = new_clipping_planes

    used_feature_combobox = field(widget_type="ComboBox", location=ImageFeaturesWidget).with_choices(FEATURE_KEYS)
    feature_editor = field(ScriptExecuteWidget, location=ImageFeaturesWidget).with_options(editor_key="minimal_surface_features")
    alpha = vfield(float, label="\u03B1", widget_type="FloatSlider", location=ImageFeaturesWidget).with_options(min=0, max=1, value=0.01)
    beta = vfield(float, label="\u03B2", location=ImageFeaturesWidget).with_options(min=.01, max=500., value=5)
    iterations = vfield(int, widget_type="Slider", location=ImageFeaturesWidget).with_options(min=1, max=10000, value=10000)
    z_scale = vfield(float, location=ImageFeaturesWidget).with_options(min=0.01, max=20., value=1.)
    use_correction = vfield(bool, location=ImageFeaturesWidget)

    blur_image_checkbox = field(bool, label="Blur image", location=BlurringWidget)
    blurring_type_combobox = (field(str, widget_type="ComboBox", location=BlurringWidget)
                              .with_choices([GAUSSIAN_BLURRING, CAD_BLURRING]))
    blur_sigma_slider = field(float, widget_type="FloatSlider", location=BlurringWidget).with_options(max=20)
    conductance_spinbox = field(float, location=BlurringWidget).with_options(max=100, value=9)
    cad_iterations_spinbox = field(int, label="Iterations", location=BlurringWidget).with_options(min=1, max=20, value=5)

    slice_annotation_method = vfield(str, location=SliceAnnotationWidget).with_choices(EstimationWorker.ANNOTATION_METHODS)

    def __init__(self):
        self._collapsible_group = CollapsibleContainerGroup()
        self._viewer = None
        self._prev_labels = None
        self.blurred_layer = None
        self._try_blurring_button = None
        self.blurring_progress_widget = ProgressWidget(message="blurring...")
        self.estimation_progress_widget = ProgressWidget(message="Annotating...")
        self.points_layer = None
        self._estimation_running = False
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)
        # Parameters for the estimator

        # self.z_scale_spinbox.valueChanged.connect(self) TODO

        # Blurring parameters
        # delayed_updater.processing.connect(
        #     lambda _: self.blurring_progress_widget.setVisible(blur_demo_btn.isChecked())) # TODO
        # delayed_updater.processed.connect(lambda _: self.blurring_progress_widget.setVisible(False)) # TODO

        self.crop_mask = None

        # self.use_meeting_points_checkbox = QCheckBox("Slice from meeting points") #TODO check if needed
        # self.use_meeting_points_checkbox.setChecked(True)
        # self.slice_widget.clipping_widget.add_layer(self.points_layer)

    def __post_init__(self):
        self.call_button.clicked.connect(self._start_estimation)
        print(self.native.children()[1].sizeHint().width())
        print(self.native.children()[1].verticalScrollBar().sizeHint().width())
        print(self.native.children()[1].widget().layout().getContentsMargins()[2])
        self.native.setMinimumWidth(self.native.children()[1].widget().sizeHint().width()
                                    + self.native.children()[1].verticalScrollBar().sizeHint().width()
                                    + 2 * self.native.children()[1].widget().layout().getContentsMargins()[2])
        for container in [self.ImageFeaturesWidget, self.BlurringWidget, self.SliceAnnotationWidget]:
            container._widget._expand_btn.setChecked(False)
            # container._widget._inner_qwidget.setStyleSheet("padding: 0 10 0 10")
            container.collapsed = True
            self._collapsible_group.addItem(container)
        self._on_feature_changed(self.used_feature)
        self.try_blurring_button.native.setCheckable(True)
        self._on_blur_checked(self.blur_image_checkbox.value)
        self._on_labels_changed()
        self._ensure_3d_image()

    def __magicclass_serialize__(self):
        d = serialize(self)
        if "image_layer_combobox" in d:
            del d["image_layer_combobox"]
        if "labels_layer_combobox" in d:
            del d["labels_layer_combobox"]
        return d

    @property
    def viewer(self):
        return self._viewer

    @property
    def image_layer(self):
        return self.image_layer_combobox.value

    @image_layer.setter
    def image_layer(self, new_layer):
        self.image_layer_combobox.value = new_layer

    @property
    def labels_layer(self):
        return self.labels_layer_combobox.value

    @labels_layer.setter
    def labels_layer(self, new_layer):
        self.labels_layer_combobox.value = new_layer

    @property
    def is_blurring_enabled(self):
        return self.blur_image_checkbox.value

    @property
    def used_feature(self):
        return self.used_feature_combobox.value

    @property
    def blurring_type(self):
        return self.blurring_type_combobox.value

    @property
    def blurring_conductance(self):
        return self.conductance_spinbox.value

    @property
    def blurring_iterations(self):
        return self.cad_iterations_spinbox.value

    @property
    def try_blurring_button(self):
        if self._try_blurring_button is None:
            for w in self._list:
                if w.name == "try_blurring":
                    self._try_blurring_button = w
                    break
        return self._try_blurring_button

    def _initialize(self, viewer: napari.Viewer, minimal_contour_widget: MinimalContourWidget):
        self._viewer = viewer
        self.minimal_contour_widget = minimal_contour_widget
        self.points_layer = self.viewer.add_points(ndim=3, size=2, name="3D Anchors [DO NOT REMOVE]")
        self.points_layer.events.connect(ColorPairsCallback())
        self.points_layer.mouse_drag_callbacks.append(self._drag_points_callback)
        self.viewer.layers.events.inserted.connect(keep_layer_on_top(self.points_layer))
        self.viewer.layers.events.moved.connect(keep_layer_on_top(self.points_layer))
        self.ImageSliceWidget.viewer = self.viewer
        self.ImageSliceWidget.clipping_widget = self.ClippingWidget
        self._ensure_3d_image()
        self._set_slice_widget_image()
        self._update_custom_feature_image()
        self.ImageSliceWidget.add_mouse_drag_callback(self._on_slice_clicked)
        self.ClippingWidget.viewer = self.viewer

    @set_design(text="Try", location=BlurringWidget,)
    def try_blurring(self):
        self._toggle_blur_demo()

    @labels_layer_combobox.connect
    def _on_labels_changed(self, idx=None):
        if self._prev_labels == self.labels_layer:
            return
        self.ImageSliceWidget.clipping_widget.remove_layer(self._prev_labels)
        self.ImageSliceWidget.clipping_widget.add_layer(self.labels_layer)
        self._prev_labels = self.labels_layer

    @used_feature_combobox.connect
    def _on_feature_changed(self, feature):
        self.feature_editor.visible = feature == "Custom"
        correct_container_size(self.ImageFeaturesWidget)

    def _add_image(self, layer_name, data, layer_args):
        if 'opacity' not in layer_args:
            layer_args['opacity'] = 0.3
        if 'blending' not in layer_args:
            layer_args['blending'] = "additive"
        if 'rgb' not in layer_args:
            layer_args['rgb'] = False
        self._viewer.add_layer(Image(data, name=layer_name, **layer_args))

    def _mask_received(self, data, offset):
        offset = offset.round().astype(int)
        bb_slice = tuple(slice(offs, offs + extent) for offs, extent in zip(offset, data.shape))
        self.labels_layer.data[bb_slice][data > 0] = self.labels_layer.data.max()+1
        self.labels_layer.refresh()

    def _refresh_layer(self, layer_name):
        if layer_name not in self._viewer.layers:
            return
        layer = self._viewer.layers[layer_name]
        # layer.data = layer.data
        layer.refresh()

    def _data_remover(self, name):
        self._viewer.layers.remove(name)

    def _show_estimation_progress(self, n_objects):
        self.estimation_progress_widget.setMaximum(n_objects)
        self.estimation_progress_widget.setValue(0)
        self.estimation_progress_widget.setVisible(True)

    def _run_feature_extractor(self, image):
        self.feature_editor.variables["image"] = image
        res = dict()

        def store_features(result_dict):
            res.update(result_dict)
        self.feature_editor.script_worker.done.connect(store_features, Qt.DirectConnection)
        self.feature_editor.Run()
        self.feature_editor.script_thread.wait()
        if "exception" in res:
            raise res["exception"]
        return res["features"]

    # def hook_callbacks(self, estimator, stage, layer_name, layer_params, data_idx=None):
    #
    #     def initializer(arr, idx):
    #         self.data_initializer(layer_name, data_idx, layer_params)(arr, idx)
    #
    #     estimator.hook_stage_data_init_event(stage, initializer)
    #     estimator.hook_stage_iteration_event(stage, self.data_updater(layer_name))
    #     estimator.hook_stage_finished_event(stage, self.data_finalizer(layer_name))

    def _drag_points_callback(self, layer, event):
        if layer.mode != "select":
            return
        yield
        if event.type == "mouse_move":
            if len(layer.selected_data) > 2 or len(layer.selected_data) == 0:
                return
            if len(layer.selected_data) == 2:
                selection_iter = iter(layer.selected_data)
                idx1 = next(selection_iter)
                idx2 = next(selection_iter)
                if idx1 // 2 != idx2 // 2:
                    return
            else:
                index = next(iter(layer.selected_data))
                if index % 2 == 0:
                    idx1 = index
                    idx2 = index + 1
                else:
                    idx1 = index - 1
                    idx2 = index
            p1 = layer.data[idx1]
            p2 = layer.data[idx2]
            bb = pts_2_bb(p1, p2, self.image_layer.data.shape)
            bb_slice = bb_2_slice(bb)
            image = self.image_layer.data[bb_slice]
            crop_mask = generate_sphere_mask(image.shape)
            self.image_layer.visible = False
            cropped_image = self._viewer.add_image(
                image*crop_mask,
                colormap=self.image_layer.colormap,
                translate=bb.min(0)*self.image_layer.scale,
                scale=self.image_layer.scale,
                rendering="attenuated_mip"
            )
            self._viewer.layers.selection.active = layer
        yield
        while event.type == "mouse_move":
            yield
        self._viewer.layers.remove(cropped_image)
        self._update_blurred_layer()
        self.image_layer.visible = True

    @blurring_type_combobox.connect
    def _on_blur_selection(self, blur_method):
        self.blur_sigma_slider.visible = blur_method == GAUSSIAN_BLURRING
        self.conductance_spinbox.visible = blur_method == CAD_BLURRING
        self.cad_iterations_spinbox.visible = blur_method == CAD_BLURRING
        correct_container_size(self.BlurringWidget)

    @blur_image_checkbox.connect
    def _on_blur_checked(self, is_checked):
        self.blurring_type_combobox.visible = is_checked
        correct_container_size(self.BlurringWidget)
        self._on_blur_selection(self.blurring_type_combobox.value if is_checked else "")

    def _blur_image(self, image):
        blur_type = self.blurring_type
        if blur_type == GAUSSIAN_BLURRING:
            return gaussian_filter(image, self.blur_sigma_slider.value)
        elif blur_type == CAD_BLURRING:
            sitk_image = sitk.GetImageFromArray(image.astype(float))
            blurred = sitk.CurvatureAnisotropicDiffusion(
                sitk_image,
                conductanceParameter=self.blurring_conductance,
                numberOfIterations=self.blurring_iterations
            )
            return sitk.GetArrayFromImage(blurred)
        elif blur_type != "":
            raise NotImplementedError("Unkown blur type %s" % blur_type)

    def _toggle_blur_demo(self):
        is_checked = not self.try_blurring_button.native.isChecked()
        self.try_blurring_button.native.setChecked(is_checked)
        print(is_checked)
        if len(self.points_layer.data) < 2:
            return
        image_layer = self.image_layer
        if image_layer is None:
            return
        if is_checked:
            bb = pts_2_bb(self.points_layer.data[0], self.points_layer.data[1], image_layer.data.shape, image_layer.scale)
            bb = bb.round().astype(int)
            bb_slice = bb_2_slice(bb)
            blurred_image = image_layer.data[bb_slice]
            self.crop_mask = generate_sphere_mask(blurred_image.shape)
            offset = bb.min(0)

            blurred = self._blur_image(blurred_image)
            blurred *= self.crop_mask
            self.blurred_layer = self._viewer.add_image(
                blurred,
                name="%s cropped" % image_layer.name,
                translate=np.add(image_layer.translate, offset),
                scale=image_layer.scale,
                colormap=image_layer.colormap,
                contrast_limits=image_layer.contrast_limits
            )
        else:
            if self.blurred_layer:
                self._viewer.layers.remove(self.blurred_layer)
                self.blurred_layer = None
            self.crop_mask = None

    @blurring_type_combobox.connect
    @conductance_spinbox.connect
    @cad_iterations_spinbox.connect
    @blur_sigma_slider.connect
    @delay_function
    def _update_blurred_layer(self, *args):
        if self.blurred_layer is None:
            return
        image_layer = self.image_layer
        if image_layer is None:
            return
        bb = pts_2_bb(self.points_layer.data[0], self.points_layer.data[1], image_layer.data.shape, image_layer.scale)
        bb = bb.round().astype(int)
        offset = bb.min(0)
        bb_slice = bb_2_slice(bb)
        blurred_image = image_layer.data[bb_slice]
        if self.crop_mask is None or self.crop_mask.shape != blurred_image.shape:
            self.crop_mask = generate_sphere_mask(blurred_image.shape)
        blurred = self._blur_image(blurred_image)
        blurred *= self.crop_mask
        self.blurred_layer.data = blurred
        self.blurred_layer.translate = np.add(image_layer.translate, offset)

    @image_layer_combobox.connect
    def _ensure_3d_image(self, _=None):
        if self.image_layer is None:
            return
        if self.image_layer.ndim == 2:
            warnings.warn("Minimal Surface works only with 3D images!")
        is_enabled = self.image_layer is not None and self.image_layer.ndim==3
        print("is enabled: ", is_enabled)
        for container in self._collapsible_group:
            container.enabled = is_enabled
        self.ClippingWidget.enabled = is_enabled
        self.call_button.enabled = is_enabled

    @image_layer_combobox.connect
    def _set_slice_widget_image(self, _=None):
        self.ImageSliceWidget.layer = self.image_layer if self.image_layer is None or self.image_layer.ndim == 3 else None

    @image_layer_combobox.connect
    def _update_custom_feature_image(self, _=None):
        if self.image_layer is not None and self.image_layer.ndim != 2:
            self.feature_editor.variables["image"] = self.image_layer.data

    def _on_slice_clicked(self, layer, event):
        start_point, end_point = layer.get_ray_intersections(
            position=event.position,
            view_direction=event.view_direction,
            dims_displayed=event.dims_displayed,
            world=True,
        )
        dragging = False
        point = None
        if (start_point is not None) and (end_point is not None):
            point = np.add(start_point, end_point)/2
            point = layer.data_to_world(point)
        yield
        while event.type == "mouse_move":
            dragging = True
            yield
        if dragging:
            return
        if event.button == 2:
            self.points_layer.remove_selected()
            return
        if point is not None:
            if QApplication.keyboardModifiers() & Qt.AltModifier:
                for i, slider in enumerate(self.ImageSliceWidget.clipping_widget.sliders):
                    range = slider.value
                    range_center = (range[0] + range[1])/2
                    new_center = int(point[i])
                    offset = new_center-range_center
                    slider.value = (range[0]+offset, range[1]+offset)
            else:
                self.points_layer.add(self.points_layer.world_to_data(point))


class ShortestPathsWidget(WidgetWithLayerList):
    shapes_data_received = Signal(str, "PyQt_PyObject", "PyQt_PyObject")

    def __init__(self, viewer: napari.Viewer):
        super().__init__(viewer, [("meeting_plane", Image), ("distance_map", Image), ("starting_points", Points)], "nd_annotator_shortest_paths")
        self.shapes_data_received.connect(self.add_shapes)
        self.viewer = viewer
        self.estimator = minimal_surface.MinimalSurfaceCalculator()
        layout = QVBoxLayout()

        layout.addWidget(QLabel("Extract contour from meeting plane"))
        plane_2_contour_btn = QPushButton("Extract")
        plane_2_contour_btn.clicked.connect(self.plane_2_contour)
        layout.addWidget(plane_2_contour_btn)
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setFixedHeight(1)
        line.setStyleSheet("background-color: #414851;")
        layout.addWidget(line)

        run_button = QPushButton("Show shortest paths")
        run_button.clicked.connect(self.show_shortest_paths)
        layout.addWidget(run_button)
        self.setLayout(layout)

    def add_shapes(self, layer_name, data, layer_args):
        if 'opacity' not in layer_args:
            layer_args['opacity'] = 0.3
        self.viewer.add_layer(Shapes(data, name=layer_name, shape_type="path", **layer_args))

    def plane_2_contour(self):
        if self.meeting_plane.layer is None:
            return
        meeting_plane_layer = self.meeting_plane.layer
        plane_mask = (meeting_plane_layer.data > 0).astype(np.uint8)
        contour = np.squeeze(cv2.findContours(np.squeeze(plane_mask), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[0][0])
        contour = np.concatenate([np.fliplr(contour), np.ones([len(contour), 1])], axis=1)
        contour = contour @ meeting_plane_layer.rotate.T + meeting_plane_layer.translate
        contour_layer = Points(contour, ndim=3, name="contour points", size=2)
        self.viewer.add_layer(contour_layer)
        self.starting_points.layer = contour_layer

    def calculate(self, points, distance_map, translate, rotate):
        paths = []
        for point in points:
            shortest_path = self.estimator.resolve_shortest_paths(point, distance_map).copy()
            paths.append(shortest_path)
        colors = generate_label_colors(len(paths))
        self.shapes_data_received.emit("Shortest path", paths, {
            'translate': translate,
            'rotate': rotate,
            'edge_color': np.concatenate([np.asarray(colors) / 255, np.ones([len(colors), 1])], axis=1),
            'edge_width': 0.1
        })

    def show_shortest_paths(self):
        distance_map = self.distance_map.layer
        points = self.starting_points.layer
        if distance_map is None or points is None:
            return
        translation = distance_map.translate
        rotation = distance_map.rotate
        distance_map = distance_map.data
        points = (points.data - translation).astype(int)
        self.viewer.window.worker2 = Thread(target=self.calculate, args=[points, distance_map, translation, rotation])
        self.viewer.window.worker2.start()

if minimal_surface is not None:
    import SimpleITK as sitk
    MinimalSurfaceWidget = _MinimalSurfaceWidget
else:
    MinimalSurfaceWidget = None




import warnings
from scipy.ndimage import binary_dilation, binary_erosion
from napari.layers import Labels, Image

LOCK_CHAR = u"\U0001F512"

def extend_mask(layer):
    if layer is None:
        return
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        labels = layer._slice.image.raw
    mask = labels == layer.selected_label
    mask = binary_dilation(mask)
    labels[mask] = layer.selected_label
    layer.events.data()
    layer.refresh()


def reduce_mask(layer):
    if layer is None:
        return
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        labels = layer._slice.image.raw
    mask = labels == layer.selected_label
    eroded_mask = binary_erosion(mask)
    labels[mask & ~eroded_mask] = 0
    layer.events.data()
    layer.refresh()


def increment_selected_label(layer):
    if layer is None:
        return
    layer.selected_label = layer.selected_label+1


def decrement_selected_label(layer):
    if layer is None:
        return
    layer.selected_label = max(0, layer.selected_label-1)


def scroll_to_prev(viewer):
    def scroll_to_prev(_):
        if len(viewer.dims.not_displayed) == 0:
            return
        viewer.dims.set_current_step(viewer.dims.not_displayed[0],
                                     viewer.dims.current_step[viewer.dims.not_displayed[0]]-1)
    return scroll_to_prev


def scroll_to_next(viewer):
    def scroll_to_next(_):
        if len(viewer.dims.not_displayed) == 0:
            return
        viewer.dims.set_current_step(viewer.dims.not_displayed[0],
                                     viewer.dims.current_step[viewer.dims.not_displayed[0]] + 1)
    return scroll_to_next


def increase_brush_size(layer):
    if layer is None:
        return
    diff = min(max(1, layer.brush_size // 10), 5)
    layer.brush_size = max(0, layer.brush_size + diff)


def decrease_brush_size(layer):
    if layer is None:
        return
    diff = min(max(1, layer.brush_size // 10), 5)
    layer.brush_size = max(0, layer.brush_size - diff)


def lock_layer(event):
    for layer in event.source:
        if layer.name.startswith(LOCK_CHAR):
            layer.editable = False


def keep_layer_on_top(layer):
    def on_top_callback(e):
        layer_list = e.source
        if layer not in layer_list:
            return
        with layer_list.events.moved.blocker(keep_layer_on_top):
            try:
                for i in reversed(range(len(layer_list))):
                    elem = layer_list[i]
                    if elem == layer:
                        break
                    if type(elem) not in [Labels, Image]:
                        continue
                    layer_index = layer_list.index(layer)
                    if i == layer_index+1:
                        layer_list.move(i, layer_index)
                    else:
                        layer_list.move(layer_index, i)
                    break
            except Exception as e:
                ...
    return on_top_callback



from qtpy.QtWidgets import QDialog, QTableWidget, QTableWidgetItem, QVBoxLayout, QAbstractItemView, QSizePolicy
from qtpy.QtCore import Qt
shortcuts = {
    "Ctrl+I": "Interpolate",
    "E/Shift+Wheel " + u"\u2191": "Increment selected label",
    "Q/Shift+Wheel " + u"\u2193": "Decrement selected label",
    "A/Ctrl+Wheel " + u"\u2191": "Previous slice",
    "D/Ctrl+Wheel " + u"\u2193": "Next slice",
    "W/Alt+Wheel " + u"\u2191": "Increase paint brush size",
    "S/Alt+Wheel " + u"\u2193": "Decrease paint brush size",
    "Ctrl+Tab": "Jump from \"Anchors\" layer to the labels layer, or vice versa",
    "Ctrl+[1-10]": "Jump to the layer at the selected index"
}


class HelpDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        table_widget = QTableWidget()
        table_widget.setRowCount(len(shortcuts))
        table_widget.setColumnCount(2)
        table_widget.setHorizontalHeaderLabels(["Action", "Shortcut"])
        for i, (shortcut, action) in enumerate(shortcuts.items()):
            table_widget.setItem(i, 0, QTableWidgetItem(action))
            table_widget.setItem(i, 1, QTableWidgetItem(shortcut))
        table_widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        table_widget.setFocusPolicy(Qt.NoFocus)
        table_widget.setSelectionMode(QAbstractItemView.NoSelection)
        table_widget.verticalHeader().hide()
        table_widget.resizeColumnsToContents()
        table_widget.setWordWrap(True)
        table_widget.horizontalHeader().sectionResized.connect(table_widget.resizeRowsToContents)
        layout = QVBoxLayout()
        layout.addWidget(table_widget)
        self.setLayout(layout)
        dialogWidth = table_widget.horizontalHeader().length() + 50
        dialogHeight = table_widget.verticalHeader().length() + 24
        self.setFixedSize(dialogWidth, dialogHeight)



import traceback
import warnings
import inspect
from functools import lru_cache

import numpy as np
import skimage
from PyQt5.QtCore import QMutex, Qt
from napari._vispy.overlays.base import LayerOverlayMixin, VispySceneOverlay
from napari._vispy.utils import visual
from napari.layers import Labels
from napari.layers.labels._labels_constants import Mode
from napari.layers.labels._labels_utils import mouse_event_to_labels_coordinate
from napari.settings import get_settings
from napari_nd_annotator.minimal_contour._eikonal_wrapper import MinimalContourCalculator
from vispy.scene import Polygon, Line, Compound, Markers

from napari_nd_annotator._helper_functions import layer_get_order

from .minimal_contour_overlay import MinimalContourOverlay

def bbox_around_points(pts):
    p1 = pts.min(0)
    p2 = pts.max(0)
    size = p2 - p1
    from_i = p1 - size[0]*0.1 - 10
    to_i = p2 + size[0]*0.1 + 10
    return from_i, to_i


def _only_when_enabled(callback):
    """Decorator that wraps a callback of VispyLabelsPolygonOverlay.

    It ensures that the callback is only executed when all the conditions are met:
    1) The overlay is enabled;
    2) The number of displayed dimensions is 2 (it can only work in 2D);
    3) The number of dimensions across which labels will be edited is 2.

    If 2, 3 are not met, the Labels mode is automatically switched to PAN_ZOOM.
    """
    if inspect.isgeneratorfunction(callback):
        def decorated_callback(self, layer: Labels, event):
            if not self.overlay.enabled:
                return

            if self._features_shape is None:
                return

            if layer._slice_input.ndisplay != 2 or layer.n_edit_dimensions != 2:
                return
            yield from callback(self, layer, event)
    else:
        def decorated_callback(self, layer: Labels, event):
            if not self.overlay.enabled:
                return

            if self._features_shape is None:
                return

            if layer._slice_input.ndisplay != 2 or layer.n_edit_dimensions != 2:
                return
            callback(self, layer, event)

    return decorated_callback


cache_size = 20_000
dash_len = 1
skip_len = 7
dashed_cache = np.concatenate([np.arange(i, i+dash_len, dtype=np.uint32) for i in range(0, cache_size*dash_len, dash_len+skip_len)])
dashed_cache = np.tile(dashed_cache[:, None], (1, 2))
dashed_cache[:, 1] += 4


@lru_cache(maxsize=200)
def get_dashed_cache(data_len):
    return dashed_cache[:np.searchsorted(dashed_cache[:, 1], data_len)]


class VispyMinimalContourOverlay(LayerOverlayMixin, VispySceneOverlay):
    layer: Labels
    mc_calculator = MinimalContourCalculator(None, 3)
    _features_shape = None
    def __init__(
            self, *, layer: Labels, overlay: MinimalContourOverlay, parent=None
    ):
        self._features = None
        points = [(0, 0), (1, 1)]
        self.point_triangle = np.zeros((3, 2), dtype=np.float64) - 1  # start point, current position, end point

        self._nodes_kwargs = {
            'face_color': (1, 1, 1, 1),
            'size': 8.0,
            'edge_width': 1.0,
            'edge_color': (0, 0, 0, 1),
            'symbol': 'x'
        }

        self._anchor_points = Markers(pos=np.array(points), **self._nodes_kwargs)
        self._last_anchor_to_current_pos_contour = Line(pos=points, width=overlay.contour_width, color=(1., 0., 0., 1.), method='gl')
        self._current_pos_to_first_anchor_contour = Line(pos=points, width=overlay.contour_width, color=(.3, .3, .3, 1.), method='gl')
        self._stored_contour = Line(pos=points, width=overlay.contour_width, method='agg')
        super().__init__(
            node=Compound([self._anchor_points, self._last_anchor_to_current_pos_contour, self._current_pos_to_first_anchor_contour, self._stored_contour]),
            layer=layer,
            overlay=overlay,
            parent=parent,
        )

        self.layer.mouse_move_callbacks.append(self._on_mouse_move)
        self.layer.mouse_drag_callbacks.append(self._on_mouse_press)
        self.layer.mouse_double_click_callbacks.append(
            self._on_mouse_double_click
        )

        self.overlay.events.anchor_points.connect(self._on_points_change)
        self.overlay.events.enabled.connect(self._on_enabled_change)
        self.overlay.events.use_straight_lines.connect(self._set_use_straight_lines)
        self.overlay.events.contour_width.connect(self._on_width_change)
        layer.events.selected_label.connect(self._update_color)
        layer.events.colormap.connect(self._update_color)
        layer.events.opacity.connect(self._update_color)
        layer.bind_key("Escape", self.clear, overwrite=True)
        # set completion radius based on settings

        self.reset()
        self._update_color()
        # If there are no points, it won't be visible
        self.overlay.visible = True
        self.move_mutex = QMutex()
        self.modifiers = None
        self.use_straight_lines = False


    @property
    def shift_down(self):
        return False # self.modifiers & Qt.ShiftModifier if self.modifiers is not None else False

    @property
    def alt_down(self):
        return self.modifiers & Qt.AltModifier if self.modifiers is not None else False

    @staticmethod
    def set_calculator_feature(features, grad_x=None, grad_y=None):
        if features is None:
            VispyMinimalContourOverlay._features_shape = None
            return
        if grad_x is None:
            grad_x = np.empty((0, 0))
        if grad_y is None:
            grad_y = np.empty((0, 0))
        if grad_x.shape != grad_y.shape or (grad_x.size != 0 and grad_x.shape != features.shape[:2]):
            raise ValueError("feature and gradient shapes should be equal")
        if features.ndim == 2:
            features = np.tile(features[..., None],(1, 1, 3))
        VispyMinimalContourOverlay.mc_calculator.set_image(features, grad_x, grad_y)
        VispyMinimalContourOverlay._features_shape = features.shape[:2]

    def _set_use_straight_lines(self):
        self.use_straight_lines = self.overlay.use_straight_lines

    def _on_enabled_change(self):
        if self.overlay.enabled:
            self._on_points_change()

    def _on_width_change(self):
        self._last_anchor_to_current_pos_contour.set_data(width=self.overlay.contour_width)
        self._current_pos_to_first_anchor_contour.set_data(width=self.overlay.contour_width)
        self._stored_contour.set_data(width=self.overlay.contour_width)

    def _on_points_change(self):
        current_pos_to_first_anchor_points = np.round(self.overlay.current_pos_to_first_anchor_contour).reshape(-1, self.layer.ndim)
        stored_contour_points = np.round(self.overlay.stored_contour).reshape(-1, self.layer.ndim)
        last_anchor_to_current_pos_points = np.round(self.overlay.last_anchor_to_current_pos_contour).reshape(-1, self.layer.ndim)
        # points = np.concatenate([current_pos_to_first_anchor_points, stored_contour_points, last_anchor_to_current_pos_points], axis=0)

        anchor_points = self.overlay.anchor_points

        if len(anchor_points):
            anchor_points = np.array(anchor_points)[
                     :, self._dims_displayed[::-1]
                     ]
        else:
            anchor_points = np.empty((0, 2))

        # try:
        for visual, new_data, is_dashed in [
            (self._stored_contour, stored_contour_points, False),
            (self._last_anchor_to_current_pos_contour, last_anchor_to_current_pos_points, True),
            (self._current_pos_to_first_anchor_contour, current_pos_to_first_anchor_points, True)
        ]:
            if len(new_data):
                new_data = np.array(new_data)[
                         :, self._dims_displayed[::-1]
                         ]
            else:
                new_data = np.empty((0, 2))
            if len(new_data) > 2:
                if isinstance(visual, Polygon):
                    visual.pos = new_data
                else:
                    if is_dashed:
                        visual.set_data(new_data, connect=get_dashed_cache(len(new_data)))
                    else:
                        visual.set_data(new_data)
                visual.visible = True
            else:
                visual.visible = False

        # except AssertionError:
        #     ...
        # except ValueError:
        #     ...
        # except IndexError:
        #     ...
        # except Exception:
        #     traceback.print_exc()
        self._anchor_points.set_data(
            pos=anchor_points,
            **self._nodes_kwargs,
        )

    def _set_color(self, color):
        border_color = tuple(color[:3]) + (1,)  # always opaque
        polygon_color = color

        # Clean up polygon faces before making it transparent, otherwise
        # it keeps the previous visualization of the polygon without cleaning
        # if polygon_color[-1] == 0:
        #     self._polygon.mesh.set_data(faces=[])
        # self._polygon.color = polygon_color

        # self._polygon.border_color = border_color
        self._stored_contour.set_data(color=border_color)

    def _update_color(self):
        layer = self.layer
        if layer.selected_label == layer.colormap.background_value:
            self._set_color((1, 0, 0, 0))
        else:
            self._set_color(
                layer._selected_color.tolist()[:3] + [layer.opacity]
            )

    @_only_when_enabled
    def _on_mouse_move(self, layer, event):
        """Continuously redraw the latest polygon point with the current mouse position."""
        if not self.move_mutex.tryLock():
            return
        try:
            if self._num_points == 0:
                return
            # image_layer = self.image_layer
            # if image_layer is None or layer._slice_input.ndisplay == 3:
            #     return
            dims_displayed = self._dims_displayed
            pos = self._get_mouse_coordinates(event)
            pos = np.clip(pos, 0, np.subtract(self.layer.data.shape, 1))
            self.overlay.anchor_points = self.overlay.anchor_points[:-1] + [pos.tolist()]
            self.point_triangle[1] = np.asarray(pos)[dims_displayed]
            if np.any(self.point_triangle < 0):
                return
            if self.use_straight_lines:
                results = [
                    np.asarray(skimage.draw.line(
                        round(self.point_triangle[1, 0]), round(self.point_triangle[1, 1]),
                        round(self.point_triangle[0, 0]), round(self.point_triangle[0, 1])
                    )).T,
                    np.asarray(skimage.draw.line(
                        round(self.point_triangle[2, 0]), round(self.point_triangle[2, 1]),
                        round(self.point_triangle[1, 0]), round(self.point_triangle[1, 1])
                    )).T
                ]
            else:
                if self._features_shape is None:
                    warnings.warn("Features not initialized.")
                    return
                results = self._estimate()

            new_e_data = np.tile(np.where([i in dims_displayed for i in range(self.layer.ndim)], np.nan,
                                          self.overlay.anchor_points[0]), [len(results[0]), 1])
            new_s_data = np.tile(np.where([i in dims_displayed for i in range(self.layer.ndim)], np.nan,
                                          self.overlay.anchor_points[0]), [len(results[1]), 1])

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                order = layer_get_order(self.layer)
                new_e_data[np.isnan(new_e_data)] = np.flipud(results[0][:, order[:2]]).reshape(-1)
                new_s_data[np.isnan(new_s_data)] = np.flipud(results[1][:, order[:2]]).reshape(-1)

            self.overlay.last_anchor_to_current_pos_contour = new_e_data.tolist()
            self.overlay.current_pos_to_first_anchor_contour = new_s_data.tolist()

        finally:
            self.move_mutex.unlock()

    @_only_when_enabled
    def _on_mouse_press(self, layer, event):
        pos = self._get_mouse_coordinates(event)
        dims_displayed = self._dims_displayed
        pos[dims_displayed] = np.clip(pos[dims_displayed], 0,
                                      np.subtract(self.layer.data.shape, 1)[dims_displayed])
        dragged = False
        yield
        while event.type == "mouse_move":
            dragged = True
            yield
        if dragged:
            return

        if event.button == 1:  # left mouse click

            orig_pos = pos.copy()
            # recenter the point in the center of the image pixel
            pos[dims_displayed] = np.round(pos[dims_displayed])
            #TODO handle undo

            prev_point = (
                None if self._num_points <= 1 else self.overlay.anchor_points[0] if self.shift_down else self.overlay.anchor_points[-2]
            )
            # Add a new point only if it differs from the previous one
            if prev_point is None or not np.allclose(np.round(prev_point), np.round(pos)):
                if self.shift_down:
                    self.overlay.anchor_points.insert(0, pos.tolist())
                    if len(self.overlay.current_pos_to_first_anchor_contour):
                        self.overlay.stored_contour = (self.overlay.current_pos_to_first_anchor_contour) + self.overlay.stored_contour
                        self.overlay.current_pos_to_first_anchor_contour = []
                else:
                    self.overlay.anchor_points = self.overlay.anchor_points[:-1] + [pos.tolist()]
                    if len(self.overlay.last_anchor_to_current_pos_contour):
                        self.overlay.stored_contour.extend(self.overlay.last_anchor_to_current_pos_contour)
                        self.overlay.last_anchor_to_current_pos_contour = []
                self.overlay.anchor_points.append((orig_pos + 1e-3).tolist())
                self.point_triangle[-1] = [self.overlay.anchor_points[0][d] for d in dims_displayed]
                self.point_triangle[0] = [self.overlay.anchor_points[-1][d] for d in dims_displayed]
        # TODO handle undo
        # elif event.button == 2 and self._num_points > 0:  # right mouse click
        #     if self._num_points < 3:
        #         self.overlay.anchor_points = []
        #     else:
        #         self.overlay.anchor_points = self.overlay.anchor_points[:-2] + [pos.tolist()]

    @_only_when_enabled
    def _on_mouse_double_click(self, layer, event):
        if event.button == 2:
            self._on_mouse_press(layer, event)
            return None
        self.overlay.stored_contour = self.overlay.current_pos_to_first_anchor_contour + self.overlay.stored_contour
        # Remove the last point from double click, but keep the vertex
        self.overlay.anchor_points = self.overlay.anchor_points[:-1]

        self.overlay.add_polygon_to_labels(layer)
        return None

    def _get_mouse_coordinates(self, event):
        pos = mouse_event_to_labels_coordinate(self.layer, event)
        if pos is None:
            return None

        pos = np.array(pos, dtype=float)
        # pos = np.floor(pos)+0.5
        # pos[self._dims_displayed] += 0.5

        return pos

    @property
    def _dims_displayed(self):
        return self.layer._slice_input.displayed

    @property
    def _num_points(self):
        return len(self.overlay.anchor_points)

    def reset(self):
        super().reset()
        self._on_points_change()

    def _estimate(self):
        from_i, to_i = bbox_around_points(self.point_triangle)
        from_i = np.clip(from_i, 0, np.asarray(self._features_shape)).astype(int)
        to_i = np.clip(to_i, 0, np.asarray(self._features_shape)).astype(int)
        self.mc_calculator.set_boundaries(from_i[1], from_i[0], to_i[1], to_i[0])
        results = self.mc_calculator.run(
            self.point_triangle,
            True,
            True,
            True
        )
        return results

    def clear(self, _):
        self.overlay.last_anchor_to_current_pos_contour = []
        self.overlay.current_pos_to_first_anchor_contour = []
        self.overlay.stored_contour = []
        self.overlay.anchor_points = []
        self.reset()


visual.overlay_to_visual[MinimalContourOverlay] = VispyMinimalContourOverlay


import json
import traceback
from pathlib import Path
from typing import Union, Optional
import itertools

import napari
from magicclass import magicclass, MagicTemplate, vfield, set_design, field
from magicclass.serialize import serialize, deserialize
from magicgui._util import debounce
from napari.layers.labels._labels_utils import sphere_indices
from qtpy.QtCore import QObject, QEvent, Qt
from qtpy.QtWidgets import QSizePolicy
from napari.layers import Labels, Image
from napari.qt.threading import create_worker
import numpy as np
from skimage.draw import draw
from scipy.ndimage import binary_fill_holes, find_objects
import math
from packaging import version
import warnings
from dataclasses import field as dfield

from .._napari_version import NAPARI_VERSION

if NAPARI_VERSION >= version.parse("0.4.15"):
    try:
        from napari_bbox import BoundingBoxLayer
    except ImportError:
        BoundingBoxLayer = None
else:
    BoundingBoxLayer = None

from .interpolation_widget import InterpolationWidget
from .minimal_contour_widget import MinimalContourWidget
from .minimal_surface_widget import MinimalSurfaceWidget
from ._utils.callbacks import (
    extend_mask,
    reduce_mask,
    increase_brush_size,
    decrease_brush_size,
    scroll_to_next,
    scroll_to_prev,
    increment_selected_label,
    decrement_selected_label,
    lock_layer,
    LOCK_CHAR
)
from ._utils.help_dialog import HelpDialog
from .._helper_functions import layer_ndisplay, layer_dims_displayed, layer_slice_indices, layer_dims_order, \
    _coerce_indices_for_vectorization, layer_get_order


@magicclass(name="Annotation Toolbox")
class AnnotatorWidget(MagicTemplate):
    autofill_objects = vfield(bool).with_options(tooltip="When drawing labels,"
                                              " close the drawn curve and fill its area after releasing the mouse")

    @magicclass(widget_type="tabbed")
    class ToolsWidget:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            interpolation_widget = field(InterpolationWidget, name="Interpolation")
            minimal_contour_widget = field(MinimalContourWidget, name="Minimal Contour")
            minimal_surface_widget = field(MinimalSurfaceWidget, name="Minimal Surface") if MinimalSurfaceWidget is not None else None
        persist = dfield(default=bool)

    def __init__(self, viewer: napari.Viewer, persist=True):
        self._viewer = viewer
        self.persist = persist
        self._active_labels_layer = None
        self._active_bbox_layer = None

        self._viewer.layers.selection.events.connect(self._on_layer_selection_change)
        self._viewer.layers.selection.events.connect(lock_layer)
        self._viewer.layers.events.inserted.connect(self._move_bbox_to_top)
        self._viewer.layers.events.moved.connect(self._move_bbox_to_top)

        # self.native.installEventFilter(self) # TODO fix
        self._set_fill_objects(self.autofill_objects)
        self._on_layer_selection_change()
        # self._init_bbox_layer()

    def __post_init__(self):
        if self.persist:
            try:
                self._load(quiet=True)
            except Exception as e:
                traceback.print_exc()
        self.changed.connect(self._on_change)
        for w in self._list:
            if w.name == "show_help_window":
                w.native.setFixedSize(20, 20)
                w.native.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
                self.native.layout().setAlignment(w.native, Qt.AlignRight)
        self.ToolsWidget.interpolation_widget._initialize(self._viewer, self)
        self.ToolsWidget.minimal_contour_widget._initialize(self._viewer)
        if self.ToolsWidget.minimal_surface_widget is not None:
            self.ToolsWidget.minimal_surface_widget._initialize(self._viewer, self.ToolsWidget.minimal_contour_widget)

    @debounce(wait=0.5)
    def _on_change(self, *args, **kwargs) -> None:
        if self.persist:
            self._dump()

    @property
    def _dump_path(self) -> Path:
        from magicgui._util import user_cache_dir

        name = getattr(self.__class__, "__qualname__", str(self.__class__))
        name = name.replace("<", "-").replace(">", "-")  # e.g. <locals>
        return user_cache_dir("napari") / f"{self.__class__.__module__}.{name}"

    @debounce
    def _dump(self, path: Optional[Union[str, Path]] = None) -> None:
        data = serialize(self)
        with open(path or self._dump_path, "w") as save_file:
            json.dump(data, save_file)

    def _load(self, path: Optional[Union[str, Path]] = None, quiet: bool = False) -> None:
        try:
            with open(path or self._dump_path) as save_file:
                data = json.load(save_file)
                print(data)
                deserialize(self, data)
        except Exception as e:
            if not quiet:
                raise e

    @set_design(visible=False)
    def eventFilter(self, source: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Hide:
            self._disconnect_all()
            return True
        elif event.type() == QEvent.Show:
            self._connect_all()
            return True
        return self.native.eventFilter(source, event)

    @property
    def active_labels_layer(self):
        return self._active_labels_layer

    @active_labels_layer.setter
    def active_labels_layer(self, labels_layer):
        if not isinstance(labels_layer, Labels) and labels_layer is not None:
            raise TypeError("'active_labels_layer' should be a Labels layer.")
        if labels_layer == self.active_labels_layer:
            return
        self._unset_labels_callbacks()
        self._active_labels_layer = labels_layer
        self._set_fill_objects(self.autofill_objects)
        self._set_labels_callbacks()
        # self._init_bbox_layer()

    def _unset_labels_callbacks(self):
        if self.active_labels_layer is None:
            return
        self.active_labels_layer.bind_key("Control-+", overwrite=True)(None)
        self.active_labels_layer.bind_key("Control--", overwrite=True)(None)
        self.active_labels_layer.bind_key("Q", overwrite=True)(None)
        self.active_labels_layer.bind_key("E", overwrite=True)(None)
        self.active_labels_layer.bind_key("A", overwrite=True)(None)
        self.active_labels_layer.bind_key("D", overwrite=True)(None)

    def _set_labels_callbacks(self):
        if self.active_labels_layer is None:
            return
        if self._start_update_bbox not in self.active_labels_layer.events.selected_label.callbacks:
            self.active_labels_layer.events.selected_label.connect(self._start_update_bbox)
            self.active_labels_layer.events.data.connect(self._start_update_bbox)
        self.active_labels_layer.bind_key("Control-+", overwrite=True)(extend_mask)
        self.active_labels_layer.bind_key("Control--", overwrite=True)(reduce_mask)
        self.active_labels_layer.bind_key("Q", overwrite=True)(decrement_selected_label)
        self.active_labels_layer.bind_key("E", overwrite=True)(increment_selected_label)
        self.active_labels_layer.bind_key("A", overwrite=True)(scroll_to_prev(self._viewer))
        self.active_labels_layer.bind_key("D", overwrite=True)(scroll_to_next(self._viewer))
        self.active_labels_layer.bind_key("W", overwrite=True)(increase_brush_size)
        self.active_labels_layer.bind_key("S", overwrite=True)(decrease_brush_size)

    def _move_bbox_to_top(self, e=None):
        if self._active_bbox_layer not in self._viewer.layers:
            return
        layer_list = self._viewer.layers
        with layer_list.events.moved.blocker(self._move_bbox_to_top):
            try:
                for i in reversed(range(len(layer_list))):
                    layer = layer_list[i]
                    if layer == self._active_bbox_layer:
                        break
                    if type(layer) not in [Labels, Image]:
                        continue
                    bbox_index = layer_list.index(self._active_bbox_layer)
                    if i == bbox_index+1:
                        layer_list.move(i, bbox_index)
                    else:
                        layer_list.move(bbox_index, i)
                    break
            except KeyError:
                ...

    def _start_update_bbox(self, event):
        if self._active_bbox_layer is None:
            return
        worker = create_worker(self._update_bbox, event.source)
        worker.finished.connect(self._active_bbox_layer.refresh)
        worker.start()

    def _update_bbox(self, layer):
        if self._active_bbox_layer is None:
            return
        if len(self._active_bbox_layer.data) > 0:
            self._active_bbox_layer.data = []
        label = layer.selected_label
        data = layer.data
        mask = data == label
        bb_corners = find_objects(mask, max_label=1)[0]
        if bb_corners is None:
            return
        min_ = [slice_.start for slice_ in bb_corners]
        max_ = [slice_.stop - 1 for slice_ in bb_corners]
        bb = np.asarray(np.where(list(itertools.product((False, True), repeat=layer.ndim)), max_, min_))
        self._active_bbox_layer.add(bb)
        self._active_bbox_layer.properties["label"][0] = label
        self._active_bbox_layer.refresh_text()

    @autofill_objects.connect
    def _set_fill_objects(self, state):
        if self.active_labels_layer is None:
            return
        if state and self.visible:
            if self._fill_holes not in self.active_labels_layer.mouse_drag_callbacks:
                self.active_labels_layer.mouse_drag_callbacks.append(self._fill_holes)
        elif not state:
            if self._fill_holes in self.active_labels_layer.mouse_drag_callbacks:
                self.active_labels_layer.mouse_drag_callbacks.remove(self._fill_holes)

    def _disconnect_all(self):
        if self.active_labels_layer is not None and self._fill_holes in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.remove(self._fill_holes)

    def _connect_all(self):
        if self.autofill_objects and self.active_labels_layer is not None and self._fill_holes not in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.append(self._fill_holes)

    def _on_layer_selection_change(self, event=None):
        if event is None or event.type == "changed":
            active_layer = self._viewer.layers.selection.active
            if isinstance(active_layer, Labels):
                self.active_labels_layer = active_layer
            else:
                self.active_labels_layer = None

    @staticmethod
    def _draw_line(x1, y1, x2, y2, brush_size, output):
        line_x, line_y = draw.line(x1, y1, x2, y2)
        for x, y in zip(line_x, line_y):
            cx, cy = draw.disk((x, y), np.floor(brush_size / 2) + 0.5, output.shape)
            output[cx, cy] = True

    def _fill_holes(self, layer: Labels, event):
        if layer.mode != "paint" or layer_ndisplay(layer) != 2:
            return
        coordinates = layer.world_to_data(event.position)
        poly = []
        # coordinates = tuple(max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
        dims_displayed = layer_dims_displayed(layer)
        image_coords = tuple(int(coordinates[i]) for i in range(layer.ndim) if i in dims_displayed)
        slice_dims = layer_slice_indices(layer)
        current_draw = np.zeros_like(layer.data[slice_dims], bool)
        start_x, start_y = prev_x, prev_y = image_coords
        poly.append(image_coords)
        yield
        while event.type == 'mouse_move':
            coordinates = layer.world_to_data(event.position)
            # coordinates = tuple(max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
            image_coords = tuple(int(coordinates[i]) for i in range(layer.ndim) if i in dims_displayed)
            poly.append(image_coords)
            prev_x, prev_y = image_coords
            yield
        # s = np.asarray([[0, 1, 0],
        #                 [1, 1, 1],
        #                 [0, 1, 0]])
        s = None
        coordinates = layer.world_to_data(event.position)
        # coordinates = tuple(
        #     max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
        image_coords = tuple(int(coordinates[i]) for i in range(layer.ndim) if i in dims_displayed)
        poly.append(image_coords)
        prev_x, prev_y = image_coords
        # AnnotatorWidget._draw_line(prev_x, prev_y, start_x, start_y, layer.brush_size, current_draw)
        radius = np.floor(layer.brush_size / 2) + 0.5
        lx, ly = draw.line(prev_x, prev_y, start_x, start_y)
        poly = np.asarray(poly)
        px, py = draw.polygon(poly[:, 0], poly[:, 1], current_draw.shape)
        current_draw[px, py] = True
        sphere_mask_idx = sphere_indices(radius, (1., 1.))
        for p in zip(lx, ly):
            mask_indices = sphere_mask_idx + np.round(p).astype(
                int
            )
            mask_indices = mask_indices[np.all(np.logical_and(mask_indices>=0, mask_indices<current_draw.shape), axis=1)]
            current_draw[mask_indices[:, 0], mask_indices[:, 1]] = True
        # binary_fill_holes(current_draw, output=current_draw, structure=s)
        if layer.preserve_labels:
            current_draw = current_draw & (layer.data[slice_dims] == 0)
        idx = np.nonzero(current_draw)
        change_idx = np.zeros((layer.ndim, len(idx[0])), dtype=int)
        order = layer_dims_order(layer)
        dims_displayed = layer_dims_displayed(layer)
        slice_indices = layer_slice_indices(layer)
        for i, d in enumerate(order):
            change_idx[d] = idx[layer_get_order(layer)[i - layer.ndim + 2]] if d in dims_displayed else slice_indices[d]
        change_idx = _coerce_indices_for_vectorization(layer.data, change_idx)
        if hasattr(layer, "data_setitem"):
            layer.data_setitem(change_idx, layer.selected_label)
        else:
            layer._save_history(
                (change_idx, layer.data[change_idx], layer.selected_label))
            layer.data[change_idx] = layer.selected_label
            layer.events.data()
            layer.refresh()

    @set_design(text="?", width=20, height=20)
    def show_help_window(self):
        """Show keyboard shortcuts."""
        dialog = HelpDialog(self.native)
        dialog.show()

    def _init_bbox_layer(self):
        if self.active_labels_layer is None or BoundingBoxLayer is None:
            return
        if self._active_bbox_layer in self.viewer.layers:
            if (self.active_labels_layer is not None and self._active_bbox_layer.ndim == self.active_labels_layer.ndim
                    or self.active_labels_layer is None and self._active_bbox_layer.ndim == 2):
                return
            with self.viewer.layers.selection.events.blocker(self._on_layer_selection_change):
                self.viewer.layers.remove(self._active_bbox_layer)
        self._active_bbox_layer = BoundingBoxLayer(name="%s active object" % LOCK_CHAR, face_color="transparent",
                                                  edge_color="green", ndim=self.active_labels_layer.ndim if self.active_labels_layer else 2)
        self._active_bbox_layer.opacity = 1.
        self._active_bbox_layer.current_properties |= {"label": 0}
        self._active_bbox_layer.text = {
            "text": "{label:d}",
            "size": 10,
            "color": "green"
        }
        with self.viewer.layers.selection.events.blocker(self._on_layer_selection_change):
            prev_selection = self.viewer.layers.selection.copy()
            self.viewer.add_layer(self._active_bbox_layer)
            if len(prev_selection) == 1 and self._active_bbox_layer not in prev_selection:
                self.viewer.layers.selection = prev_selection




import colorsys
import time
import warnings
from functools import wraps
import numpy as np

import skimage
from skimage.filters import gaussian

import napari
from napari.layers.labels._labels_utils import get_dtype
from napari.layers.labels.labels import _coerce_indices_for_vectorization
from napari.layers.labels._labels_constants import Mode
from napari.layers import Image, Labels
from napari.utils.action_manager import action_manager
from napari.utils._dtype import get_dtype_limits
from napari.utils.events import Event
from superqt.sliders import QDoubleRangeSlider
from napari._qt.layer_controls.qt_labels_controls import QtLabelsControls
from napari._qt.widgets.qt_mode_buttons import QtModeRadioButton, QtModePushButton
from napari._qt.qt_resources import get_current_stylesheet

from qtpy.QtCore import Signal, QObject, QEvent, QThread, Qt
from qtpy.QtWidgets import QLabel, QSizePolicy
from magicclass.serialize import serialize
from magicclass import MagicTemplate, field, abstractapi, magicclass, vfield, set_design
from magicclass.widgets import FreeWidget
from magicgui.widgets import ComboBox, Image as MagicImage

from .minimal_contour_overlay.vispy_minimal_contour_overlay import VispyMinimalContourOverlay
from .minimal_contour_overlay.minimal_contour_overlay import MinimalContourOverlay
from ._utils.changeable_color_box import QtChangeableColorBox
from ._utils import ScriptExecuteWidget, ProgressWidget
from ._utils.collapsible_widget import CollapsibleContainerGroup, correct_container_size
from ._utils.callbacks import reduce_mask, extend_mask
from .._helper_functions import layer_dims_order, layer_dims_displayed, layer_slice_indices, \
    layer_get_order
from ..minimal_contour import FeatureManager
from .._napari_version import NAPARI_VERSION
from napari_nd_annotator._widgets.resources import mc_contour_style_path, interpolate_style_path


def delay_function(function=None, delay=0.2):

    def decorator(fn):
        from threading import Timer

        _store: dict = {"timer": None, "last_call": 0.0, "args": (), "kwargs": {}}

        @wraps(fn)
        def delayed(*args, **kwargs):
            _store["args"] = args
            _store["kwargs"] = kwargs
            def call_it():
                _store["timer"] = None
                _store["last_call"] = time.time()
                return fn(*_store["args"], **_store["kwargs"])

            now = time.time()
            if not _store["last_call"] or (now - _store["last_call"]) > delay:
                ret = call_it()
                return ret
            else:
                if _store["timer"] is not None:
                    _store["timer"].cancel()
                _store["timer"] = Timer(delay, call_it)
                _store["timer"].start()
            _store["last_call"] = time.time()
            return None

        return delayed

    return decorator if function is None else decorator(function)  # type: ignore


GRADIENT_BASED = 0
INTENSITY_BASED = 2

ASSYM_GRADIENT = "Gradient [assymetric]"
SYM_GRADIENT = "Gradient [symmetric]"
HIGH_INTENSITY = "High intensity"
LOW_INTENSITY = "Low intensity"
CUSTOM = "Custom"
FEATURE_KEYS = [ASSYM_GRADIENT, SYM_GRADIENT, HIGH_INTENSITY, LOW_INTENSITY, CUSTOM]
FEATURES = {
    ASSYM_GRADIENT: GRADIENT_BASED,
    SYM_GRADIENT: INTENSITY_BASED,
    HIGH_INTENSITY: INTENSITY_BASED,
    LOW_INTENSITY: INTENSITY_BASED,
    CUSTOM: INTENSITY_BASED
}

DEMO_SIZE = 200


class ColorBox(FreeWidget):
    def __init__(self, layer: Labels = None):
        super().__init__()
        self.wdt = QtChangeableColorBox(layer)
        self.set_widget(self.wdt)

    @property
    def layer(self):
        return self.wdt.layer

    @layer.setter
    def layer(self, layer):
        self.wdt.layer = layer


class EventFilter(QObject):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def eventFilter(self, src: 'QObject', event: 'QEvent') -> bool:
        if hasattr(event, "modifiers"):
            self.widget.modifiers = event.modifiers()

        # if src == self.widget.selected_id_label and event.type() == QEvent.Enter:
        #     pos = QCursor.pos()
        # elif event.type() == QEvent.Type.HoverMove:
        #     pos = src.mapToGlobal(event.pos())
        # else:
        #     return False
        # pos.setX(pos.x()+20)
        # pos.setY(pos.y()+20)
        # self.widget.selected_id_label.move(pos)
        return False


@magicclass(name="Minimal Contour", widget_type="scrollable", properties={"x_enabled": False})
class MinimalContourWidget(MagicTemplate):
    image_layer_combobox = field(Image, label="Image")

    @magicclass(widget_type="collapsible", name="Image Features")
    class ImageFeaturesWidget(MagicTemplate):
        ...

    @magicclass(widget_type="collapsible", name="Blurring")
    class BlurWidget(MagicTemplate):
        blur_image_checkbox = abstractapi()
        blur_sigma_slider = abstractapi()

    @magicclass(widget_type="collapsible", name="Contour Options")
    class ContourWidget(MagicTemplate):
        ...

    @magicclass(widget_type="collapsible", name="Label Options")
    class LabelOptionsWidget(MagicTemplate):
        @magicclass(layout="horizontal", name="selected_label_widget", labels=False,)
        class SelectedLabelWidget(MagicTemplate):
            ...

        autoincrease_label_id = abstractapi()

        @magicclass(layout="horizontal", name="label_manipulation_widget")
        class LabelManipulationWidget(MagicTemplate):
            def reduce_mask(self): ...

            def extend_mask(self): ...


    color_box = field(ColorBox, location=LabelOptionsWidget.SelectedLabelWidget)
    selected_label_spinbox = field(int, location=LabelOptionsWidget.SelectedLabelWidget).with_options(tooltip="Selected label for the current 'labels' layer.")
    autoincrease_label_id = vfield(bool, label="Auto-increment Label", location=LabelOptionsWidget).with_options(value=True,
                                                      tooltip="When checked, selected label will be incremented\n"
                                                              "after completing the contour")

    used_feature_combobox = field(widget_type=ComboBox, label="Used Feature", options={"choices": FEATURE_KEYS}, location=ImageFeaturesWidget)
    feature_editor = field(ScriptExecuteWidget, location=ImageFeaturesWidget).with_options(editor_key="minimal_contour_features")
    param_spinbox = field(int, label="Parameter", location=ImageFeaturesWidget).with_options(min=1, max=50, value=5,
                                            tooltip="Increasing this parameter will cause contours to be more aligned\n"
                                                    "with the selected image feature (with a lower number the contour will be\n"
                                                    "closer to the Euclidean shortest path")
    local_intensity_correction_checkbox = field(bool, location=ImageFeaturesWidget).with_options(visible=False)


    blur_image_checkbox = field(bool, label="Blur image", location=BlurWidget).with_options(tooltip="When checked, the image will be blurred\n"
                                        "before calculating minimal contour")
    blur_sigma_slider = field(float, label="Sigma", widget_type="FloatSlider", location=BlurWidget).with_options(max=20, value=1)

    smooth_contour_checkbox = field(bool, label="Use contour smoothing", location=ContourWidget).with_options(
        tooltip="When checked, the finished contour will be smoothed\n"
                "to remove minor noise using Fourier transformation.")
    contour_smoothness_slider = field(float, label="Contour smoothness", widget_type="FloatSlider", location=ContourWidget).with_options(min=0., max=1.,
                                                          tooltip="Number of Fourier coefficients to approximate the contour.\n"
                                                                  "Lower number -> smoother contour\n"
                                                                  "Higher number -> more faithful to the original")
    contour_width_spinbox = field(int, label="Contour width", location=ContourWidget).with_options(min=1, max=50, value=2, tooltip="Point size for contour display.")

    def __init__(self, viewer: napari.Viewer = None):
        self.progress_dialog = ProgressWidget(self.native, message="Drawing mask...")
        self.draw_worker = self.DrawWorker()
        self.draw_thread = QThread()
        self.draw_worker.moveToThread(self.draw_thread)
        self.draw_worker.done.connect(self.draw_thread.quit)
        self.draw_worker.done.connect(self._set_mask)
        self.draw_worker.done.connect(lambda: self.progress_dialog.setVisible(False))
        self.draw_thread.started.connect(self.draw_worker.run)
        self._img = None
        self._features = None
        self._viewer: napari.Viewer = viewer
        self._labels_layer: Labels = None
        self.apply_contrast_limits = True
        self._prev_img_layer = self.image_layer
        self._orig_image = self.image_layer.data if self.image_layer else None
        self.point_triangle = np.zeros((3, 2), dtype=np.float64) - 1  # start point, current position, end point
        self.remove_last_anchor = False
        self.last_added_with_shift = None
        self.last_segment_length = None
        self.prev_n_anchor_points = 0
        self.prev_labels_layer = self.labels_layer
        self.feature_inverted = False
        self.prev_timer_id = None
        self.feature_manager = None
        self.event_filter = EventFilter(self)

        # Ctrl+Z handling

        self.selected_id_label = QLabel(parent=self.native)
        self.selected_id_label.setFixedSize(20, 20)
        self.selected_id_label.setAlignment(Qt.AlignCenter)
        self.selected_id_label.setSizePolicy(QSizePolicy.Fixed, self.selected_id_label.sizePolicy().verticalPolicy())

        self.selected_id_label.setWindowFlags(Qt.Window
                                              | Qt.FramelessWindowHint
                                              | Qt.WindowStaysOnTopHint)
        self.selected_id_label.setAttribute(Qt.WA_ShowWithoutActivating)
        self.selected_id_label.installEventFilter(self.event_filter)
        self._collapsible_group = CollapsibleContainerGroup()

        self.demo_image_widget = MagicImage()
        self.demo_image_widget.native.setFixedSize(DEMO_SIZE, DEMO_SIZE)
        self.demo_image_widget.native.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.demo_image_widget.native.setWindowFlags(Qt.Window
                                              | Qt.FramelessWindowHint
                                              | Qt.WindowStaysOnTopHint)
        self.demo_image_widget.native.setAttribute(Qt.WA_ShowWithoutActivating)

        self.modifiers = None

    def __post_init__(self):
        if self._viewer is not None:
            self._initialize(self._viewer)
        VispyMinimalContourOverlay.mc_calculator.set_param(self.param)
        self.local_intensity_correction_checkbox.changed.connect(VispyMinimalContourOverlay.mc_calculator.set_use_local_maximum)

        if self.labels_layer is not None:
            dtype_lims = get_dtype_limits(get_dtype(self.labels_layer))
        else:
            dtype_lims = 0, np.iinfo(int).max
        selected_label_spinbox = self.selected_label_spinbox
        selected_label_spinbox.min, selected_label_spinbox.max = dtype_lims
        self._set_use_smoothing(self.is_blurring_enabled, update_image=False)
        self._on_feature_change(self.used_feature, update_image=False)
        self._on_param_change(self.param, update_image=False)
        self.feature_editor.script_worker.done.connect(self._set_features)

        self.native.setMinimumWidth(self.native.children()[1].widget().sizeHint().width()
                                    + self.native.children()[1].verticalScrollBar().sizeHint().width()
                                    + 2 * self.native.children()[1].widget().layout().getContentsMargins()[2])
        for container in [self.ImageFeaturesWidget, self.BlurWidget, self.ContourWidget, self.LabelOptionsWidget]:
            container._widget._expand_btn.setChecked(False)
            # container._widget._inner_qwidget.setStyleSheet("padding: 0 10 0 10")
            # container._widget._inner_qwidget.setSizePolicy(QSizePolicy.Minimum, container._widget._inner_qwidget.sizePolicy().verticalPolicy())
            container.collapsed = True
            self._collapsible_group.addItem(container)
        self.feature_editor.visible = self.used_feature == CUSTOM
        self.feature_editor.run_button.setEnabled(self.image_layer is not None)
        self.feature_editor.run_button.setToolTip("Open an image first" if self.image_layer is None else "")
        self.param_spinbox.visible = self.used_feature == ASSYM_GRADIENT
        self.demo_image_widget.native.setFixedSize(DEMO_SIZE, DEMO_SIZE)
        self.demo_image_widget.native.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.contour_smoothness_slider.visible = self.is_contour_smoothing_enabled
        self.selected_label_spinbox.native.setKeyboardTracking(False)
        self.selected_label_spinbox.native.setAlignment(Qt.AlignCenter)
        self.blur_sigma_slider.native.children()[0].sliderPressed.connect(self._show_demo_image)
        self.blur_sigma_slider.native.children()[0].sliderReleased.connect(self._hide_demo_image)
        self.blur_sigma_slider.native.children()[0].sliderReleased.connect(self._apply_blurring)
        self.blur_sigma_slider.native.children()[0].sliderReleased.connect(self._set_image)
        self._on_label_change()
        self._on_selected_label_change()

    def __magicclass_serialize__(self):
        d = serialize(self)
        if "image_layer_combobox" in d:
            del d["image_layer_combobox"]
        if "labels_layer_combobox" in d:
            del d["labels_layer_combobox"]
        return d

    @property
    def viewer(self):
        return self._viewer

    @property
    def image_layer(self):
        return self.image_layer_combobox.value

    @image_layer.setter
    def image_layer(self, new_layer):
        self.image_layer_combobox.value = new_layer

    @property
    def image_data(self):
        return self._img

    @property
    def labels_layer(self) -> Labels:
        return self._labels_layer

    @property
    def blur_sigma(self):
        return self.blur_sigma_slider.value

    @property
    def used_feature(self):
        return self.used_feature_combobox.value

    @property
    def param(self):
        return self.param_spinbox.value

    @property
    def is_blurring_enabled(self):
        return self.blur_image_checkbox.value

    @property
    def is_contour_smoothing_enabled(self):
        return self.smooth_contour_checkbox.value

    @property
    def contour_smoothness(self):
        return self.contour_smoothness_slider.value

    @property
    def contour_width(self):
        return self.contour_width_spinbox.value

    @property
    def shift_down(self):
        return self.modifiers & Qt.ShiftModifier if self.modifiers is not None else False

    @property
    def ctrl_down(self):
        return self.modifiers & Qt.ControlModifier if self.modifiers is not None else False

    @property
    def alt_down(self):
        return self.modifiers & Qt.AltModifier if self.modifiers is not None else False

    def _initialize_helper_layers(self):
        self._change_contour_width(self.contour_width)

    class DrawWorker(QObject):
        done = Signal("PyQt_PyObject")
        contour: np.ndarray
        mask_shape: tuple

        def run(self):
            mask = skimage.draw.polygon2mask(self.mask_shape, self.contour)
            mask = skimage.filters.rank.median(mask.astype(np.uint8), np.asarray([[0, 1, 0], [1, 1, 1], [0, 1, 0]])).astype(bool)
            self.done.emit(mask)

    def _set_mask(self, mask):

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            idx = np.nonzero(mask)
            labels_layer = self.labels_layer
            ndim = labels_layer.data.ndim
            change_idx = np.zeros((ndim, len(idx[0])), dtype=int)
            order = layer_dims_order(labels_layer)
            dims_displayed = layer_dims_displayed(labels_layer)
            slice_indices = layer_slice_indices(labels_layer)
            for i, d in enumerate(order):
                change_idx[d] = idx[i-ndim+2] if d in dims_displayed else slice_indices[d]
            # change_idx = [idx[i] if i in dims_displayed else slice_indices[i] for i in range(ndim)]
            change_idx = _coerce_indices_for_vectorization(labels_layer.data, change_idx)
            if hasattr(labels_layer, "data_setitem"):
                labels_layer.data_setitem(change_idx, labels_layer.selected_label)
            else:
                labels_layer._save_history(
                    (change_idx, labels_layer._slice.image.raw[mask], labels_layer.selected_label))
                labels_layer._slice.image.raw[mask] = labels_layer.selected_label
                labels_layer.events.data()
                labels_layer.refresh()
        if labels_layer and self.autoincrease_label_id:
            labels_layer.selected_label += 1

    def _apply_blurring(self):
        image_layer = self.image_layer
        if image_layer is None or self.viewer is None:
            return
        image_layer.data = self._orig_image.copy()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.is_blurring_enabled and self.blur_sigma > 0.:
                if image_layer.data.ndim-(1 if image_layer.rgb else 0)>=3:
                    slice_ = tuple(
                        slice(None) if i in layer_dims_displayed(self.image_layer) else self.viewer.dims.current_step[i]
                        for i in range(self.viewer.dims.ndim)
                    )
                else:
                    slice_ = (slice(None), slice(None))
                try:
                    orig_image = self._orig_image[slice_]
                except IndexError:
                    warnings.warn("Current slice is out of bounds, skipping blurring")
                    return
                if image_layer.rgb:
                    max_vals = orig_image.max((0, 1))
                    for i in range(3):
                        if max_vals[i] == 0:
                            continue
                        image_layer.data[slice_ + (i,)] = self._blur_image(orig_image[..., i]).astype(self._orig_image.dtype)
                else:
                    image_layer.data[slice_] = self._blur_image(self._orig_image[slice_]).astype(self._orig_image.dtype)
                image_layer.events.data()
                image_layer.refresh()

    # tooltip = "Reduce the selected label by one pixel around the borders\n"
    # "using erosion"
    @set_design(location=LabelOptionsWidget.LabelManipulationWidget)
    def reduce_mask(self):
        if self.labels_layer is not None:
            reduce_mask(self.labels_layer)

    # tooltip = "Extend the selected label by one pixel around the borders\n"
    # "using dilation"
    @set_design(location=LabelOptionsWidget.LabelManipulationWidget)
    def extend_mask(self):
        if self.labels_layer is not None:
            extend_mask(self.labels_layer)

    @image_layer_combobox.connect
    def _on_image_changed(self, *args):
        self._store_orig_image(*args)
        self.feature_editor.run_button.setEnabled(self.image_layer is not None)
        self.feature_editor.run_button.setToolTip("Open an image first" if self.image_layer is None else "")
        self._set_image("_on_image_changed")

    # @labels_layer_combobox.connect
    def _on_label_change(self, _=None):
        labels_layer = self.labels_layer
        selected_label_spinbox = self.selected_label_spinbox
        colorbox = self.color_box
        if labels_layer is not None:
            dtype_lims = get_dtype_limits(get_dtype(labels_layer))
            selected_label_spinbox.value = labels_layer.selected_label
        else:
            dtype_lims = 0, np.iinfo(int).max
        selected_label_spinbox.min, selected_label_spinbox.max = dtype_lims
        colorbox.layer = labels_layer
        if self.prev_labels_layer is not None:
            self.prev_labels_layer.events.selected_label.disconnect(self._on_selected_label_change)
            if self._on_mouse_wheel in self.prev_labels_layer.mouse_wheel_callbacks:
                self.prev_labels_layer.mouse_wheel_callbacks.remove(self._on_mouse_wheel)
            # self.prev_labels_layer.bind_key("Shift", overwrite=True)(None)
            self.prev_labels_layer.bind_key("Alt", overwrite=True)(None)
            self.prev_labels_layer.bind_key("Control-+", overwrite=True)(None)
            self.prev_labels_layer.bind_key("Control--", overwrite=True)(None)
        if labels_layer is not None:
            labels_layer.events.selected_label.connect(
                self._on_selected_label_change
            )
            if self._on_mouse_wheel not in labels_layer.mouse_wheel_callbacks:
                labels_layer.mouse_wheel_callbacks.append(self._on_mouse_wheel)
            labels_layer.bind_key("Shift", overwrite=True)(self._shift_pressed)
            self._on_selected_label_change()
        self.prev_labels_layer = labels_layer

    @used_feature_combobox.connect
    def _on_feature_change(self, used_feature, update_image=True):
        self.feature_editor.visible = (used_feature == CUSTOM)
        self.feature_inverted = (used_feature == LOW_INTENSITY)
        VispyMinimalContourOverlay.mc_calculator.set_method(FEATURES[used_feature])
        self.param_spinbox.visible = used_feature == ASSYM_GRADIENT
        if update_image:
            self._set_image("_on_feature_change")

    @param_spinbox.connect
    def _on_param_change(self, val, update_image=True):
        VispyMinimalContourOverlay.mc_calculator.set_param(val)
        if update_image:
            self._set_image("_on_param_change")

    @blur_image_checkbox.connect
    def _set_use_smoothing(self, _, update_image=True):
        if self.viewer is None:
            return
        self._apply_blurring()
        if update_image:
            self._set_image("_set_use_smoothing")

    def _show_demo_image(self):
        pos = self.blur_sigma_slider.native.pos()
        pos.setX(min(pos.x(), self.native.size().width()-(DEMO_SIZE+10)))
        pos.setY(pos.y() + DEMO_SIZE)
        pos = self.native.mapToGlobal(pos)
        self.demo_image_widget.native.move(pos)
        self.demo_image_widget.visible = True

    def _hide_demo_image(self):
        self.demo_image_widget.visible = False

    @blur_sigma_slider.connect
    def _update_demo_image(self, event=None):
        im_layer: Image = self.image_layer
        if im_layer is None:
            img = np.ones((DEMO_SIZE, DEMO_SIZE, 3), np.uint8)
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # img = im_layer._data_view.astype(float)
                im_layer._set_view_slice()
                slice_indices = layer_slice_indices(im_layer)
                img = self._orig_image[slice_indices].astype(float)
            if all(s > DEMO_SIZE for s in (img.shape[:2] if im_layer.rgb else img.shape)):
                img = img[(img.shape[0]-DEMO_SIZE)//2:(img.shape[0]+DEMO_SIZE)//2, (img.shape[0]-DEMO_SIZE)//2:(img.shape[0]+DEMO_SIZE)//2]
            if self.is_blurring_enabled:
                img = self._blur_image(img)
            if not im_layer.rgb:
                max_ = im_layer.contrast_limits[1]
                min_ = im_layer.contrast_limits[0]
                img = np.clip((img - min_) / (max_ - min_), 0, 1)
                img = (im_layer.colormap.map(img.ravel()).reshape(img.shape + (4,))*255)[..., :3].copy()
            img = img.astype(np.uint8)
        self.demo_image_widget.value = img

    @smooth_contour_checkbox.connect
    def _on_smooth_checked(self, is_checked):
        self.contour_smoothness_slider.visible = is_checked
        correct_container_size(self.ContourWidget)
        if self.labels_layer is not None:
            self.labels_layer._overlays["minimal_contour"].contour_smoothness = self.contour_smoothness if is_checked else 1.

    @contour_smoothness_slider.connect
    def _on_smoothness_changed(self, value):
        if self.labels_layer is not None:
            self.labels_layer._overlays["minimal_contour"].contour_smoothness = value

    @contour_width_spinbox.connect
    def _change_contour_width(self, value):
        if self.labels_layer is not None:
            self.labels_layer._overlays["minimal_contour"].contour_width = value

    @selected_label_spinbox.connect
    def _change_selected_label(self, value):
        if self.labels_layer is None:
            if self.selected_label_spinbox.value != 0:
                self.selected_label_spinbox.value = 0
            return
        self.labels_layer.selected_label = value
        self.LabelOptionsWidget.SelectedLabelWidget.native.clearFocus()
        self.native.setFocus()

    def _on_selected_label_change(self):
        """Receive layer model label selection change event and update spinbox."""
        labels_layer = self.labels_layer
        if labels_layer is None:
            return
        with labels_layer.events.selected_label.blocker():
            value = labels_layer.selected_label
            self.selected_label_spinbox.value = value
            self._update_label_tooltip()

    def _shift_pressed(self, *_):
        # TODO handle correctly
        return
        # with warnings.catch_warnings():
        #     warnings.simplefilter("ignore")
        #     if self.labels_layer is not None and len(self.anchor_points.data) == 0 and self.viewer.window.qt_viewer.canvas.native.hasFocus():
        #         self.selected_id_label.show()
        # self.from_e_points_layer.face_color = "gray"
        # self.from_e_points_layer.current_face_color = "gray"
        # self.to_s_points_layer.face_color = "red"
        # self.to_s_points_layer.current_face_color = "red"
        # self.from_e_points_layer.edge_color = "gray"
        # self.from_e_points_layer.current_edge_color = "gray"
        # self.to_s_points_layer.edge_color = "red"
        # self.to_s_points_layer.current_edge_color = "red"
        # yield
        # self.selected_id_label.hide()
        # self.from_e_points_layer.face_color = "red"
        # self.from_e_points_layer.current_face_color = "red"
        # self.to_s_points_layer.face_color = "gray"
        # self.to_s_points_layer.current_face_color = "gray"
        # self.from_e_points_layer.edge_color = "red"
        # self.from_e_points_layer.current_edge_color = "red"
        # self.to_s_points_layer.edge_color = "gray"
        # self.to_s_points_layer.current_edge_color = "gray"

    def _ctrl_shift_pressed(self, _):
        shift_callback = self._shift_pressed(_)
        yield from shift_callback
        yield from shift_callback

    def _ctrl_z_callback(self, _):
        if self.labels_layer is not None and hasattr(self.labels_layer, "data_setitem"):
            self.labels_layer.undo()

    def _on_mouse_wheel(self, _, event):
        labels_layer = self.labels_layer
        if labels_layer is None:
            return
        delta = (np.sign(event.delta)).astype(int)
        if self.shift_down:
            labels_layer.selected_label = max(0, labels_layer.selected_label + delta[1])
        elif self.alt_down:
            diff = delta[0]
            diff *= min(max(1, labels_layer.brush_size//10), 5)
            labels_layer.brush_size = max(1, labels_layer.brush_size + diff)

    def _on_right_click(self, layer, event: Event):
        # TODO remove
        return
        if event.button != 2 or layer.mode != Mode.ADD:
            return
        self.remove_last_anchor = True
        if self.last_segment_length is None:
            return
        with self.anchor_points.events.data.blocker():
            if self.last_added_with_shift:
                self.anchor_points.data = self.anchor_points.data[1:]
                self.output.data = self.output.data[self.last_segment_length:]
            else:
                self.anchor_points.data = self.anchor_points.data[:-1]
                self.output.data = self.output.data[:-self.last_segment_length]
        self.point_triangle[-1] = self.anchor_points.data[0][list(layer_dims_displayed(self.anchor_points))]
        self.point_triangle[0] = self.anchor_points.data[-1][list(layer_dims_displayed(self.anchor_points))]
        self.last_segment_length = None

    def _blur_image(self, img):
        return gaussian(
            img,
            sigma=self.blur_sigma,
            preserve_range=True,
            channel_axis=-1 if img.ndim == 3 else None,
            truncate=2.
        )

    def _store_orig_image(self, *_):
        image_layer: Image = self.image_layer
        if image_layer != self._prev_img_layer:
            if self._prev_img_layer is not None:
                self._prev_img_layer.data = self._orig_image
            self._orig_image = image_layer.data.copy() if image_layer is not None else None
        self._prev_img_layer = image_layer
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            qt_viewer = self.viewer.window.qt_viewer
        if image_layer in qt_viewer.controls.widgets:
            controls = qt_viewer.controls.widgets[image_layer]
            for child in controls.children():
                if type(child) == QDoubleRangeSlider:
                    try:
                        child.sliderReleased.connect(self._set_image, Qt.UniqueConnection)
                    except TypeError:
                        pass
                    break

    def _set_image(self, event=None):
        image_layer: Image = self.image_layer
        if image_layer is None:
            self._img = None
            self.feature_editor.variables["image"] = None
            VispyMinimalContourOverlay.set_calculator_feature(None)
            # self.feature_editor.features = None #TODO check
            return
        if self.viewer is None:
            return
        if self.viewer.dims.ndisplay == 3:
            # self.anchor_points.editable = False
            return
        self._apply_blurring()
        if not image_layer.visible:
            image_layer.set_view_slice()
        if NAPARI_VERSION < "0.5.0":
            image_layer._slice_dims(self.viewer.dims.point, self.viewer.dims.ndisplay, self.viewer.dims.order)
        else:
            image_layer._slice_dims(self.viewer.dims)
        self.autoincrease_label_id = (image_layer.ndim == 2)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            image = image_layer._data_view.astype(float)
        if image.ndim == 2:
            image = np.concatenate([image[..., np.newaxis]] * 3, axis=2)
        elif image.shape[-1] == 4:
            image = image[..., :3]
        min_, max_ = image_layer.contrast_limits if self.apply_contrast_limits else (image.min(), image.max())
        image = np.clip((image - min_) / (max_ - min_), 0, 1)
        if self.feature_inverted:
            image = 1 - image
        self._img = image
        if self.is_blurring_enabled:
            image = self._blur_image(image).astype(float)
        if self.used_feature == CUSTOM:
            self.feature_editor.variables["image"] = image
            self.feature_editor.Run()
        else:
            grad_x, grad_y = self.feature_manager.get_features(image_layer)
            if grad_x is not None:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    order = layer_get_order(image_layer)
                    grad_x = grad_x.transpose(order[:2])
                    grad_y = grad_y.transpose(order[:2])
                if self.is_blurring_enabled:
                    grad_x = self._blur_image(grad_x)
                    grad_y = self._blur_image(grad_y)
                if self.used_feature == SYM_GRADIENT:
                    grad_x, grad_y = grad_x / (max_ - min_), grad_y / (max_ - min_)
                    grad_magnitude = np.linalg.norm([grad_x, grad_y], axis=0)
                    min_, max_ = grad_magnitude.min(), grad_magnitude.max()
                    grad_magnitude = (grad_magnitude - min_) / (max_ - min_)
                    VispyMinimalContourOverlay.set_calculator_feature(grad_magnitude)
                else:
                    VispyMinimalContourOverlay.set_calculator_feature(image, grad_x, grad_y)
        # self.anchor_points.translate = image_layer.translate
        # self.from_e_points_layer.translate = image_layer.translate
        # self.to_s_points_layer.translate = image_layer.translate
        # self.output.translate = image_layer.translate

    @delay_function(delay=.5)
    def _delayed_set_image(self, *args):
        self._set_image(args)

    def _set_features(self, var_dict: dict):
        if "exception" in var_dict:
            raise var_dict["exception"]
        features = var_dict.get("features", None)
        if features is None:
            warnings.warn("The 'features' variable was not set in the script.")
            return
        if features.ndim == 2:
            features = np.concatenate([features[..., np.newaxis]] * 3, -1)
        self._features = features
        VispyMinimalContourOverlay.set_calculator_feature(features.astype(float))

    def _update_label_tooltip(self):
        labels_layer = self.labels_layer
        if labels_layer is None:
            return
        self.selected_id_label.setText(str(labels_layer.selected_label))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            bg_color = labels_layer._selected_color
        if bg_color is None:
            bg_color = np.ones(4)
        h, s, v = colorsys.rgb_to_hsv(*bg_color[:-1])
        bg_color = (255*bg_color).astype(int)
        color = (np.zeros(3) if v > 0.7 else np.ones(3)*255).astype(int)
        self.selected_id_label.setStyleSheet("background-color: rgb(%d,%d,%d); color: rgb(%d,%d,%d)" % (tuple(bg_color[:-1]) + tuple(color)))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.canvas.native.setFocus()

    def _on_active_layer_changed(self, *_):
        if self.viewer is None:
            return
        active_layer = self.viewer.layers.selection.active
        self._labels_layer = active_layer if isinstance(active_layer, Labels) else None
        if self._labels_layer:
            if "minimal_contour" not in self._labels_layer._overlays:
                self._labels_layer.bind_key("Control", self._on_ctrl_pressed)
                self._labels_layer._overlays.update({"minimal_contour": MinimalContourOverlay()})
                labels_control: QtLabelsControls = self.viewer.window.qt_viewer.controls.widgets[self._labels_layer]
                mc_btn = QtModeRadioButton(self._labels_layer, "minimal contour", Mode.PAN_ZOOM)
                action_manager.bind_button(
                    'napari-nD-annotator:activate_labels_mc_mode',
                    mc_btn,
                    # extra_tooltip_text=extra_tooltip_text,
                )
                for button in labels_control.button_group.buttons():
                    button.toggled.connect(self._disable_mc_mode)
                mc_btn.setStyleSheet(get_current_stylesheet([mc_contour_style_path]))
                labels_control.button_group.addButton(mc_btn)
                labels_control._EDIT_BUTTONS += (mc_btn,)
                labels_control.mc_button = mc_btn
                labels_control.button_grid.addWidget(labels_control.mc_button, 1, 0)

                def switch_to_mc_mode(*_, **__):
                    mc_btn.blockSignals(True)
                    mc_btn.setChecked(True)
                    mc_btn.blockSignals(False)
                    self._enable_mc_mode()
                    # btn.setChecked(True)
                self._labels_layer.bind_key("0", switch_to_mc_mode)
            self.labels_layer._overlays["minimal_contour"].contour_smoothness = self.contour_smoothness if self.is_contour_smoothing_enabled else 1.

    def _on_ctrl_pressed(self, _):
        if self.labels_layer is None:
            return
        self.labels_layer._overlays["minimal_contour"].use_straight_lines = True
        yield
        self.labels_layer._overlays["minimal_contour"].use_straight_lines = False


    def _set_viewer_callbacks(self):
        self.viewer.layers.events.connect(self.image_layer_combobox.reset_choices)
        self.viewer.layers.selection.events.active.connect(self._on_active_layer_changed)

        def change_layer_callback(num):
            def change_layer(_):
                if num - 1 < len(self.viewer.layers):
                    self.viewer.layers.selection.select_only(self.viewer.layers[num - 1])
                else:
                    warnings.warn("%d is out of range (number of layers: %d)" % (num, len(self.viewer.layers)))

            return change_layer

        for i in range(1, 10):
            self.viewer.bind_key("Control-%d" % i, overwrite=True)(change_layer_callback(i))
        self.viewer.dims.events.current_step.connect(self._delayed_set_image, position="last")

        self.viewer.dims.events.ndisplay.connect(self._set_image, position="last")
        self.viewer.dims.events.order.connect(self._set_image, position="last")
        self.viewer.layers.events.removed.connect(lambda e: self.feature_manager.remove_features(e.value))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.window().installEventFilter(self.event_filter)
            self.viewer.window.qt_viewer.window().setMouseTracking(True)

    def _initialize(self, viewer: napari.Viewer):
        self._viewer = viewer
        self._initialize_helper_layers()
        self.feature_manager = FeatureManager(self.viewer)
        self._set_viewer_callbacks()
        self._store_orig_image()
        self._apply_blurring()
        self._update_label_tooltip()
        self._set_image()
        action_manager.register_action("napari-nD-annotator:activate_labels_mc_mode", self._enable_mc_mode,
                                       "We're switching to MC mode", None)

    def _enable_mc_mode(self, *args, **kwargs):
        if self.labels_layer is not None:
            self.labels_layer._overlays["minimal_contour"].enabled = True

    def _disable_mc_mode(self, state):
        if state and self.labels_layer is not None:
            self.labels_layer._overlays["minimal_contour"].enabled = False
